"""IO"""
from StringIO import StringIO


def df_to_s3_bucket(df, bucket_conn, key, **kwargs):
    """Write dataframe to s3 bucket

    :param df: data frame
    :param bucket_conn: boto3 bucket object
    :param key: object key
    :param kwargs: additional keyword arguments to pandas.to_csv
    """
    mem_file = StringIO()
    df.to_csv(mem_file, **kwargs)
    bucket_conn.put_object(Body=mem_file.getvalue(), Key=key)
    mem_file.close()



"""Data Preparation

Calculates and joins parameters required for pygypsy to a plot table

"""
#pylint: disable=no-member
from __future__ import division
import logging
import pandas as pd
from copy import deepcopy

from pygypsy.stand_density_factor import (
    estimate_sdf_aw,
    estimate_sdf_sb,
    estimate_sdf_sw,
    estimate_sdf_pl,
)
from pygypsy.density import (
    estimate_density_aw,
    estimate_density_sw,
    estimate_density_sb,
    estimate_density_pl,
)
from pygypsy.site_index import (
    get_site_indices_from_dominant_species,
    _estimate_site_index
)
from pygypsy.utils import (
    _get_gypsy_valid_species,
    _log_loop_progress,
    _generate_fplot_dict,
    _reclassify_and_sort_species,
)
from pygypsy.asaCompileAgeGivenSpSiHt import (
    computeTreeAge,
    ComputeGypsyTreeHeightGivenSiteIndexAndTotalAge,
    ComputeGypsySiteIndex,
)
from pygypsy.exceptions import MinimumAgeError, ProportionsSumError


LOGGER = logging.getLogger(__name__)


def _populate_species_dict_with_indices(species_dict, estimated_site_indices):
    """Fill the dictionary with estimated SIs

    Fill all the SIs to avoid IFs and loops. Some of them will not be
    used.

    :param species_dict: dictionary used to store params for all species
    :param estimated_site_indices: array of estimated site indices from
                                   get_species_site_indices function

    ..note: the dominant species in the estiamted_site_indices uses its
    observed site index

    """
    local_species_dict = deepcopy(species_dict)

    for species, site_index in estimated_site_indices.items():
        local_species_dict[species]['SI'] = site_index

    return local_species_dict


# TODO: this is a bit deep in the stack to use the pandas data structure
# should refactor in future. instead pass tph and bph as args
def populate_species_dict(partial_species_list,
                          sorted_species_abbrev_perc_tuples_list,
                          dominant_species=None, row=None,
                          dominant_species_current_age=None,
                          dominant_species_current_height=None):
    """Fill partial fplot

    Given fplot with proportion filled out, add age, top height, basal area,
    density

    :param partial_species_list:
    :param sorted_species_abbrev_perc_tuples_list:
    :param dominant_species:
    :param dominant_species_current_age:
    :param dominant_species_current_height:

    """
    species_dict = deepcopy(partial_species_list)
    outer_sorted_species_perc_list = sorted_species_abbrev_perc_tuples_list
    for species in outer_sorted_species_perc_list:
        # TODO: raise NoDominantSpeciesError and catch it where called instead of break
        # confirmed with julianno
        if species[1] == 0:
            break
        # TODO: this is always true on first pass of the loop
        elif species[0] == dominant_species[0]:
            species_dict[species[0]]['tage'] = dominant_species_current_age
            species_dict[species[0]]['topHeight'] = dominant_species_current_height
            species_dict[species[0]]['N'] = row['TPH'] * species[1] / 100
            species_dict[species[0]]['BA'] = row['BPH'] * species[1] / 100
            x_si = ComputeGypsySiteIndex(
                species[0],
                species_dict[species[0]]['topHeight'],
                0,
                species_dict[species[0]]['tage']
            )
            species_dict[species[0]]['bhage'] = x_si[0]
            # if, after re-arranging the proportions, dom species is another
            # one then we need to re-estimate everything  even for the new
            # dominant species
        else:
            si_sp = species_dict[species[0]]['SI']
            species_dict[species[0]]['PCT'] = species[1]
            species_dict[species[0]]['tage'] = computeTreeAge(
                species[0],
                treeHt=dominant_species_current_height,
                treeSi=si_sp,
                maxTreeAge=450
            )
            # density based on the proportion of the species
            species_dict[species[0]]['N'] = row['TPH'] * species[1]/100
            # Basal area from the species proportion as well
            species_dict[species[0]]['BA'] = row['BPH'] * species[1]/100
            # calling the ComputeGypsySiteIndex function, estimate bhage
            x_si = ComputeGypsySiteIndex(
                species[0], dominant_species_current_height, 0,
                species_dict[species[0]]['tage']
            )
            species_dict[species[0]]['bhage'] = x_si[0]
    return species_dict


def _prep_row(row, minimum_age=25):
    '''Estimate properties required for gypsy simulation

    :param row: row of data frame
    :param minimum age: age below which plot should be skipped

    :raises ProportionsSumError: if species proportions do not add to 1
    :raises MinimumAgeError: if plot is less than minimum age

    '''
    species_abbrev_percent_list = [
        (row['SP1'], row['PCT1']),
        (row['SP2'], row['PCT2']),
        (row['SP3'], row['PCT3']),
        (row['SP4'], row['PCT4']),
        (row['SP5'], row['PCT5'])
    ]
    plot_dominant_species = row['SP1']
    dominant_species_current_age = row['AGE']
    dominant_species_current_height = row['HD']

    check_prop = sum(zip(*species_abbrev_percent_list)[1])
    if check_prop != 100:
        raise ProportionsSumError(
            'Species proportions do not add to 1: %s' % check_prop
        )

    if dominant_species_current_age < minimum_age:
        msg = ('Dominant species age is less than minimum age %s;'
               'skipping plot %s') % (minimum_age, row['id_l1'])
        raise MinimumAgeError(msg)

    site_index = _estimate_site_index(
        plot_dominant_species,
        dominant_species_current_age,
        dominant_species_current_height
    )

    temp_dominant_species = _get_gypsy_valid_species(plot_dominant_species)
    gypsy_site_indices = get_site_indices_from_dominant_species(
        temp_dominant_species,
        site_index
    )
    empty_species_dict = _generate_fplot_dict()
    species_dict = _populate_species_dict_with_indices(empty_species_dict,
                                                       gypsy_site_indices)

    outer_sorted_species_perc_list, outer_species_perc_dict = \
        _reclassify_and_sort_species(species_abbrev_percent_list)

    species_dict['Aw']['PCT'] = outer_species_perc_dict['Aw']
    species_dict['Pl']['PCT'] = outer_species_perc_dict['Pl']
    species_dict['Sw']['PCT'] = outer_species_perc_dict['Sw']
    species_dict['Sb']['PCT'] = outer_species_perc_dict['Sb']

    dominant_species = outer_sorted_species_perc_list[0]

    species_dict = populate_species_dict(
        species_dict, outer_sorted_species_perc_list,
        dominant_species=dominant_species,
        row=row,
        dominant_species_current_age=dominant_species_current_age,
        dominant_species_current_height=dominant_species_current_height
    )

    density_aw = species_dict['Aw']['N']
    density_sw = species_dict['Sw']['N']
    density_pl = species_dict['Pl']['N']
    density_sb = species_dict['Sb']['N']
    tage_aw = species_dict['Aw']['tage']
    tage_sw = species_dict['Sw']['tage']
    tage_pl = species_dict['Pl']['tage']
    tage_sb = species_dict['Sb']['tage']
    bhage_aw = species_dict['Aw']['bhage']
    bhage_sw = species_dict['Sw']['bhage']
    bhage_pl = species_dict['Pl']['bhage']
    bhage_sb = species_dict['Sb']['bhage']
    y2bh_aw = tage_aw - bhage_aw
    y2bh_sw = tage_sw - bhage_sw
    y2bh_pl = tage_pl - bhage_pl
    y2bh_sb = tage_sb - bhage_sb
    site_index_aw = species_dict['Aw']['SI']
    site_index_sw = species_dict['Sw']['SI']
    site_index_pl = species_dict['Pl']['SI']
    site_index_sb = species_dict['Sb']['SI']
    basal_area_aw = species_dict['Aw']['BA']
    basal_area_sb = species_dict['Sb']['BA']
    basal_area_sw = species_dict['Sw']['BA']
    basal_area_pl = species_dict['Pl']['BA']

    top_height_aw = ComputeGypsyTreeHeightGivenSiteIndexAndTotalAge(
        'Aw', site_index_aw, tage_aw
    )
    top_height_sb = ComputeGypsyTreeHeightGivenSiteIndexAndTotalAge(
        'Sb', site_index_sb, tage_sb
    )
    top_height_pl = ComputeGypsyTreeHeightGivenSiteIndexAndTotalAge(
        'Pl', site_index_pl, tage_pl
    )
    top_height_sw = ComputeGypsyTreeHeightGivenSiteIndexAndTotalAge(
        'Sw', site_index_sw, tage_sw
    )
    density_bh_aw, sdf_aw0 = estimate_sdf_aw(
        'Aw', site_index_aw, bhage_aw, density_aw
    )
    density_bh_sb, sdf_sb0 = estimate_sdf_sb(
        'Sb', site_index_sb, tage_sb, density_sb
    )
    density_bh_sw, sdf_sw0 = estimate_sdf_sw(
        'Sw', site_index_sw, tage_sw, sdf_aw0, density_sw
    )
    density_bh_pl, sdf_pl0 = estimate_sdf_pl(
        'Pl', site_index_pl, tage_pl,
        sdf_aw0, sdf_sw0, sdf_sb0, density_pl
    )

    initial_density_aw = estimate_density_aw(
        sdf_aw0, 0, site_index_aw
    )
    initial_density_sb = estimate_density_sb(
        sdf_sb0, 0, site_index_sb
    )
    initial_density_sw = estimate_density_sw(
        sdf_sw0, sdf_aw0, 0, site_index_sw
    )
    initial_density_pl = estimate_density_pl(
        sdf_aw0, sdf_sw0, sdf_sb0, sdf_pl0, 0, site_index_pl
    )

    plot_properties = {
        'SI_Aw': site_index_aw,
        'SI_Sw': site_index_sw,
        'SI_Pl': site_index_pl,
        'SI_Sb': site_index_sb,
        'N_Aw': density_aw,
        'N_Sw': density_sw,
        'N_Pl': density_pl,
        'N_Sb': density_sb,
        'y2bh_Aw': y2bh_aw,
        'y2bh_Sw': y2bh_sw,
        'y2bh_Pl': y2bh_pl,
        'y2bh_Sb': y2bh_sb,
        'tage_Aw': tage_aw,
        'tage_Sw': tage_sw,
        'tage_Pl': tage_pl,
        'tage_Sb': tage_sb,
        'BA_Aw': basal_area_aw,
        'BA_Sw': basal_area_sw,
        'BA_Pl': basal_area_pl,
        'BA_Sb': basal_area_sb,
        'SDF_Aw': sdf_aw0,
        'SDF_Sw': sdf_sw0,
        'SDF_Pl': sdf_pl0,
        'SDF_Sb': sdf_sb0,
        'N0_Aw': initial_density_aw,
        'N0_Sb': initial_density_sb,
        'N0_Sw': initial_density_sw,
        'N0_Pl': initial_density_pl,
        'StumpDOB_Aw': species_dict['Aw']['StumpDOB'],
        'StumpDOB_Sb': species_dict['Sb']['StumpDOB'],
        'StumpDOB_Sw': species_dict['Sw']['StumpDOB'],
        'StumpDOB_Pl': species_dict['Pl']['StumpDOB'],
        'StumpHeight_Aw': species_dict['Aw']['StumpHeight'],
        'StumpHeight_Sb': species_dict['Sb']['StumpHeight'],
        'Stumpheight_Sw': species_dict['Sw']['StumpHeight'],
        'StumpHeight_Pl': species_dict['Pl']['StumpHeight'],
        'TopDib_Aw': species_dict['Aw']['TopDib'],
        'TopDib_Sb': species_dict['Sb']['TopDib'],
        'TopDib_Sw': species_dict['Sw']['TopDib'],
        'TopDib_Pl': species_dict['Pl']['TopDib'],
        'topHeight_Aw': top_height_aw,
        'topHeight_Sw': top_height_sw,
        'topHeight_Sb': top_height_sb,
        'topHeight_Pl': top_height_pl
    }

    return plot_properties


def prep_standtable(data, **kwargs):
    '''Define site_index of all other species given the dominant species

    The site index is only defined for the dominant species in a plot. This
    function uses the dominant species and its site index to estimate the site
    indices for other species. It estimates other values and returns a data
    frame of all of the values for all plots input.

    estimates site_index Dom from inventory Age and HD
    this site_index is used to estimate the site index of the other species
    for example:
    - FD is dom species take HD and Age
    - assume FD is Sw generate site_index for Sw
    - get the SW site_index
    - calculate the SIs for other species from the conversion formulas

    :param data: input data frame
    :param kwargs: additional arguments to :func:`_prep_row`
    '''
    plot_dict = {}
    n_rows = data.shape[0]

    for i, row in data.iterrows():
        _log_loop_progress(i, n_rows)
        plot_id = row['id_l1']

        try:
            prepped_plot = _prep_row(row, **kwargs)
            plot_dict[plot_id] = prepped_plot
        except MinimumAgeError as err:
            LOGGER.error(err.message)
        except ProportionsSumError as err:
            LOGGER.error(err.message)
        except Exception as err: #pylint: disable=broad-except
            LOGGER.critical('Unhandled error: %s', err.message)

    plot_df = pd.DataFrame(plot_dict)
    plot_df = plot_df.transpose()

    return plot_df



"""Basal Area Simulation"""
#pylint: disable=no-member
# these  functions can be improved because the increment is not
# autoregressive, so we can calculate it as a vector operation using
# the other arrays then the actual values are just the cumulative sums
# TODO: change initial age to data age
# TODO: change present density to density_at_data

import logging
import numpy as np

from pygypsy import basal_area_increment as incr


LOGGER = logging.getLogger(__name__)


# TODO: make this consistent with the other functions will need an option for
# making species proportion constant and not expentially increasing with the factor
def sim_basal_area_aw(initial_age, site_index, density_at_bh_age,
                      basal_area_at_bh_age, sdf_aw, correction_factor, densities,
                      use_correction_factor_future=False,
                      stop_at_initial_age=True,
                      fix_proportion_and_density_to_initial_age=False,
                      species_proportion_at_bh_age=None, present_density=None,
                      force_use_densities=False):
    '''Simlulate basal area forward in time for White Aspen

    It creates the trajectory of basal area from bhage up to the inventory year
    given a correction factor that is being optimized

    :param float initial_age: Clock that uses the oldest species as a reference
                              to become the stand age
    :param float site_index: site index of species Aw
    :param float basal_area_at_bh_age: basal area of Aw at breast height age
    :param float SDF_Aw0: Stand Density Factor of species Aw
    :param float density_at_bh_age: initial density of species Aw at breast height age
    :param float correction_factor: correction factor that guarantees that trajectory
                                    passes through data obtained with inventory
    :param bool use_correction_factor_future: switch that determine whether the correction
                                              factor will be used for the future years
    :param bool stop_at_initial_age: switch that determines whether simulation
                                     will stop at the date of the inventory or
                                     will continue for the length of the densities object
    :param bool fix_proportion_and_density_to_initial_age: if true, uses the
    proportion and density from the initial age for all years of the
    simulation. this is provided for consistency with a previous version which
    did this implicityly in for estiamting the basal area factors
    :param bool force_use_densities: ignore other parameters and just use them                                     densities object for calculating basal
                                     area increment

    '''
    max_age = initial_age if stop_at_initial_age else len(densities)

    basal_area_aw_arr = np.zeros(max_age)
    basal_area_temp = basal_area_at_bh_age

    for i, spec_comp_dict in enumerate(densities[0: max_age]):
        sc_factor = correction_factor \
                    if i < initial_age or use_correction_factor_future \
                       else 1
        if force_use_densities:
            years_from_bhage = spec_comp_dict['bhage_Aw']
            spec_proportion = spec_comp_dict['SC_Aw']
            present_density = spec_comp_dict['N_bh_AwT']
        else:
            years_from_bhage = spec_comp_dict['bhage_Aw']
            # TODO: revise to match the funcs of the other species
            spec_proportion = densities[initial_age]['SC_Aw'] \
                              if fix_proportion_and_density_to_initial_age \
                                 else spec_comp_dict['SC_Aw']
            present_density = densities[initial_age]['N_bh_AwT'] \
                              if fix_proportion_and_density_to_initial_age \
                                 else spec_comp_dict['N_bh_AwT']

        if density_at_bh_age > 0:
            if years_from_bhage > 0:
                spec_proportion = spec_proportion * sc_factor
                basal_area_increment = incr.increment_basal_area_aw(
                    spec_proportion, site_index, present_density,
                    density_at_bh_age, years_from_bhage, basal_area_temp
                )
                basal_area_temp = basal_area_temp + basal_area_increment
                new_basal_area = basal_area_temp

                if new_basal_area < 0:
                    new_basal_area = 0
            else:
                new_basal_area = 0
        else:
            basal_area_temp = 0
            new_basal_area = 0

        basal_area_aw_arr[i] = new_basal_area

    return basal_area_aw_arr


def sim_basal_area_sb(initial_age, site_index, density_at_bh_age,
                      basal_area_at_bh_age, correction_factor, densities,
                      use_correction_factor_future=False,
                      stop_at_initial_age=True,
                      fix_proportion_and_density_to_initial_age=False,
                      species_proportion_at_bh_age=None, present_density=None,
                      force_use_densities=False):
    '''Simlulate basal area forward in time for Black Spruce

    It creates the trajectory of basal area from bhage up to the inventory year
    given a correction factor that is being optimized

    :param float initial_age: Clock that uses the oldest species as a reference to become
                              the stand age
    :param float initial_age_sb: species specific age counted independently
    :param float site_index: site index of species Sb
    :param float basal_area_at_bh_age: basal area of Sb at breast height age
    :param float density_at_bh_age: initial density of species Sb at breast height age
    :param float correction_factor: correction factor that guarantees that trajectory
                                    passes through data obtained with inventory
    :param bool use_correction_factor_future: switch that determine whether the correction
                                              factor will be used for the future years
    :param bool stop_at_initial_age: switch that determines whether simulation
                                     will stop at the date of the inventory or
                                     will continue for the length of the densities object

    :param bool fix_proportion_and_density_to_initial_age: if true, uses the
    proportion and density from the initial age for all years of the
    simulation. this is provided for consistency with a previous version which
    did this implicityly in for estiamting the basal area factors
    :param bool force_use_densities: ignore other parameters and just use them                                     densities object for calculating basal
                                     area increment

    '''
    max_age = initial_age if stop_at_initial_age else len(densities)
    basal_area_arr = np.zeros(max_age)
    basal_area_temp = basal_area_at_bh_age

    for i, spec_comp_dict in enumerate(densities[0: max_age]):
        sc_factor = correction_factor \
                    if i < initial_age or use_correction_factor_future \
                       else 1
        if force_use_densities:
            years_from_bhage = spec_comp_dict['bhage_Aw']
            spec_proportion = spec_comp_dict['SC_Aw']
            present_density = spec_comp_dict['N_bh_AwT']
        else:
            years_from_bhage = spec_comp_dict['bhage_Sb']
            if i==0:
                spec_proportion = species_proportion_at_bh_age
            elif i < initial_age:
                pass
            else:
                spec_proportion = spec_comp_dict['SC_Sw']
            present_density = present_density \
                              if i < initial_age or fix_proportion_and_density_to_initial_age \
                              else spec_comp_dict['N_bh_SbT']

        if density_at_bh_age > 0:
            if years_from_bhage > 0:
                spec_proportion = spec_proportion * sc_factor
                basal_area_increment = incr.increment_basal_area_sb(
                    spec_proportion, site_index, present_density,
                    density_at_bh_age, years_from_bhage, basal_area_temp
                )
                basal_area_temp = basal_area_temp + basal_area_increment
                new_basal_area = basal_area_temp

                if new_basal_area < 0:
                    new_basal_area = 0
            else:
                new_basal_area = 0
        else:
            basal_area_temp = 0
            new_basal_area = 0

        basal_area_arr[i] = new_basal_area

    return basal_area_arr


def sim_basal_area_sw(initial_age, site_index, density_at_bh_age, sdf_aw,
                      sdf_pl, sdf_sb, basal_area_at_bh_age, correction_factor,
                      densities, use_correction_factor_future=False,
                      stop_at_initial_age=True,
                      fix_proportion_and_density_to_initial_age=False,
                      species_proportion_at_bh_age=None, present_density=None,
                      force_use_densities=False):
    '''Simlulate basal area forward in time for White Spruce
    It created the trajectory of basal area from bhage up to the inventory year
    given a correction factor that is being optimized

    :param float initial_age: Clock that uses the oldest species as a reference to
                              become the stand age
    :param float site_index: site index of species Sw
    :param float basal_area_at_bh_age: basal area of Sw at breast height age
    :param float density_at_bh_age: initial density of species Sw at breast height age
    :param float sdf_pl: Stand Density Factor of species Pl
    :param float sdf_aw: Stand Density Factor of species Aw
    :param float sdf_sb: Stand Density Factor of species Sb
    :param float correction_factor: correction factor that guarantees that trajectory
                                    passes through data obtained with inventory
    :param bool use_correction_factor_future: switch that determine whether the correction
                                              factor will be used for the future years
    :param bool stop_at_initial_age: switch that determines whether simulation
                                     will stop at the date of the inventory or
                                     will continue for the length of the densities object
    :param bool fix_proportion_and_density_to_initial_age: if true, uses the
    proportion and density from the initial age for all years of the
    simulation. this is provided for consistency with a previous version which
    did this implicityly in for estiamting the basal area factors
    :param bool force_use_densities: ignore other parameters and just use them                                     densities object for calculating basal
                                     area increment

    '''
    max_age = initial_age if stop_at_initial_age else len(densities)
    basal_area_arr = np.zeros(max_age)
    basal_area_temp = basal_area_at_bh_age

    # some of the conditional assignments in this loop could be simplified by
    # initializing them outside the loop and assigning them when the simulation
    # passes the initial age. this is not done because we are striving to reach
    # a place where these are more configurable and thus easier to test
    for i, spec_comp_dict in enumerate(densities[0: max_age]):
        sc_factor = correction_factor \
                    if i < initial_age or use_correction_factor_future\
                       else 1
        if force_use_densities:
            years_from_bhage = spec_comp_dict['bhage_Aw']
            spec_proportion = spec_comp_dict['SC_Aw']
            present_density = spec_comp_dict['N_bh_AwT']
        else:
            years_from_bhage = spec_comp_dict['bhage_Sw']
            # the first time in this for loop, spec proportion must be assigned
            if i==0:
                spec_proportion = species_proportion_at_bh_age
            # until the year of the data is reached we then do not reassign it, and the
            # value of spec_proportion * sc_factor increases exponentially
            elif i < initial_age:
                pass
            # future values use the estimated species proportion and it is constant
            else:
                spec_proportion = spec_comp_dict['SC_Sw']
            present_density = present_density \
                              if i < initial_age or fix_proportion_and_density_to_initial_age \
                              else spec_comp_dict['N_bh_SwT']

        if density_at_bh_age > 0:
            if years_from_bhage > 0:
                spec_proportion = spec_proportion * sc_factor
                basal_area_increment = incr.increment_basal_area_sw(
                    spec_proportion, site_index, present_density,
                    density_at_bh_age, years_from_bhage, sdf_aw, sdf_pl,
                    sdf_sb, basal_area_temp
                )
                basal_area_temp = basal_area_temp + basal_area_increment
                new_basal_area = basal_area_temp

                if new_basal_area < 0:
                    new_basal_area = 0
            else:
                new_basal_area = 0
        else:
            basal_area_temp = 0
            new_basal_area = 0

        basal_area_arr[i] = new_basal_area

    return basal_area_arr


def sim_basal_area_pl(initial_age, site_index, density_at_bh_age, sdf_aw,
                      sdf_sw, sdf_sb, basal_area_at_bh_age, correction_factor,
                      densities, use_correction_factor_future=False,
                      stop_at_initial_age=True,
                      fix_proportion_and_density_to_initial_age=False,
                      species_proportion_at_bh_age=None, present_density=None,
                      force_use_densities=False):
    '''Simlulate basal area forward in time for Lodgepole Pine

    :param float initial_age: Clock that uses the oldest species as a reference to
                              become the stand age
    :param float site_index: site index of species Pl
    :param float basal_area_at_bh_age: basal area of Pl at breast height age
    :param float density_at_bh_age: initial density of species Pl at breast height age
    :param float sdf_sw: Stand Density Factor of species Sw
    :param float sdf_aw: Stand Density Factor of species Aw
    :param float sdf_sb: Stand Density Factor of species Sb
    :param float correction_factor: correction factor that guarantees that trajectory
                                    passes through data obtained with inventory
    :param bool use_correction_factor_future: switch that determine whether the correction
                                              factor will be used for the future years
    :param bool stop_at_initial_age: switch that determines whether simulation
                                     will stop at the date of the inventory or
                                     will continue for the length of the densities object
    :param bool fix_proportion_and_density_to_initial_age: if true, uses the
    proportion and density from the initial age for all years of the
    simulation. this is provided for consistency with a previous version which
    did this implicityly in for estiamting the basal area factors
    :param bool force_use_densities: ignore other parameters and just use them                                     densities object for calculating basal
                                     area increment

    '''
    max_age = initial_age if stop_at_initial_age else len(densities)
    basal_area_arr = np.zeros(max_age)
    basal_area_temp = basal_area_at_bh_age

    for i, spec_comp_dict in enumerate(densities[0: max_age]):
        sc_factor = correction_factor \
                    if i < initial_age or use_correction_factor_future \
                       else 1
        if force_use_densities:
            years_from_bhage = spec_comp_dict['bhage_Aw']
            spec_proportion = spec_comp_dict['SC_Aw']
            present_density = spec_comp_dict['N_bh_AwT']
        else:
            years_from_bhage = spec_comp_dict['bhage_Pl']
            # the first time in this for loop, spec proportion must be assigned
            if i==0:
                spec_proportion = species_proportion_at_bh_age
            # until the year of the data is reached we then do not reassign it, and the
            # value of spec_proportion * sc_factor increases exponentially
            elif i < initial_age:
                pass
            # future values use the estimated species proportion and it is constant
            else:
                spec_proportion = spec_comp_dict['SC_Pl']
            present_density = present_density \
                              if i < initial_age or fix_proportion_and_density_to_initial_age \
                              else spec_comp_dict['N_bh_PlT']

        if density_at_bh_age > 0:
            if years_from_bhage > 0:
                # factor empirically determined to work better when multiplied
                # with whole increment
                basal_area_increment = sc_factor \
                                       * incr.increment_basal_area_pl(
                                           spec_proportion, site_index,
                                           present_density, density_at_bh_age,
                                           years_from_bhage, sdf_aw, sdf_sw,
                                           sdf_sb, basal_area_temp
                                       )
                basal_area_temp = basal_area_temp + basal_area_increment
                new_basal_area = basal_area_temp

                if new_basal_area < 0:
                    new_basal_area = 0
            else:
                new_basal_area = 0
        else:
            basal_area_temp = 0
            new_basal_area = 0

        basal_area_arr[i] = new_basal_area

    return basal_area_arr



"""Basal Area correction factor estimators"""
# TODO: validate that none of the variables taken from kwargs are missing
import logging
import numpy as np

from pygypsy.basal_area_simulate import (
    sim_basal_area_aw,
    sim_basal_area_sw,
    sim_basal_area_sb,
    sim_basal_area_pl,
)


LOGGER = logging.getLogger(__name__)


def get_basal_area_factors_for_all_species(**kwargs):
    f_sb = 0
    f_aw = 0
    f_sw = 0
    f_pl = 0

    if kwargs['N0_Aw'] > 0:
        f_aw = estimate_basal_area_factor_aw(**kwargs)
    if kwargs['N0_Sb'] > 0:
        f_sb = estimate_basal_area_factor_sb(**kwargs)
    if kwargs['N0_Sw'] > 0:
        f_sw = estimate_basal_area_factor_sw(**kwargs)
    if kwargs['N0_Pl'] > 0:
        f_pl = estimate_basal_area_factor_pl(**kwargs)

    return {
        'f_Aw':f_aw,
        'f_Sb':f_sb,
        'f_Sw':f_sw,
        'f_Pl':f_pl,
    }


def estimate_basal_area_factor_aw(**kwargs):
    '''This function guarantees that the trajectory of the species basal area
    passes through the basal area measured when the data was collected
    (inventory data)

    There is a trade-off between the precision one wants this estimate, which is
    given by the parameter - acceptableDiff -, and the time the convergence time

    :param float startTage: Clock that uses the oldest species as a reference to become
                            the stand age
    :param float SI_bh_Aw: site index of species Aw
    :param float N_bh_AwT: density of sp Aw at time T (it varies over time)
    :param float BA_Aw0: basal area of Aw at breast height age, assumed to be very small
    :param float BA_AwT: basal area of Aw at time T
    :param float SDF_Aw0: Stand Density Factor of species Aw
    :param float N0_Aw: initial density of species Aw at breast height age
    :param list densities: array of 'densities' objects
    '''
    age_at_observation = kwargs['startTage']
    site_index = kwargs['SI_bh_Aw']
    present_density = kwargs['N_bh_AwT']
    basal_area_at_bh_age = kwargs['BA_Aw0']
    present_basal_area = kwargs['BA_AwT']
    sdf_aw = kwargs['SDF_Aw0']
    density_at_bh_age = kwargs['N0_Aw']
    densities = kwargs['densities']
    factor = 100
    factor1 = 100 * factor
    tolerance = 0.01 * present_basal_area
    within_tolerance = False
    iter_count = 0

    while not within_tolerance:
        ba_est = sim_basal_area_aw(age_at_observation, site_index, density_at_bh_age,
                                   basal_area_at_bh_age, sdf_aw, factor,
                                   densities, use_correction_factor_future=False,
                                   stop_at_initial_age=True,
                                   fix_proportion_and_density_to_initial_age=False,
                                   species_proportion_at_bh_age=None,
                                   present_density=present_density)[-1]

        if abs(present_basal_area - ba_est) < tolerance:
            within_tolerance = True
        else:
            if (present_basal_area - ba_est) < 0:
                factor1 = factor
                factor_p = factor \
                           * (1 \
                              + (np.log10(present_basal_area) - np.log10(abs(ba_est))) \
                              / (100*np.log10(abs(ba_est))))
                factor = (factor_p+factor)/2
            elif (present_basal_area - ba_est) > 0:
                factor = (factor+factor1)/2

        iter_count += 1

        if iter_count == 10000:
            LOGGER.warning(('Slow convergence with Basal Area: %s'
                            ' and factor:%s '), ba_est, factor)
            break
    return factor


def estimate_basal_area_factor_sb(**kwargs):
    '''Black spruce basal area factor estimator

    This function guarantees that the trajectory of the species basal area
    passes through the basal area measured when the data was collected
    (inventory data)

    :param float startTage: Clock that uses the oldest species as a reference to
                            become the stand age
    :param float SI_bh_Sb: site index of species Sb
    :param float N_bh_SbT: density of sp Sb at time T (it varies over time)
    :param float BA_Sb0: basal area of Sb at breast height age, assumed to be very small
    :param float BA_SbT: basal area of Sb at time T
    :param float N0_Sb: initial density of species Sb at breast height age
    :param float y2bh_Sb: time elapseed in years from zero to breast height age of sp Sb
    :param float startTageSb: species specific ages counted independently
    :param float SC_Sb: proportion of species Sb in the stand
    :param list densities: array of 'densities' objects

    '''
    age_at_observation = kwargs['startTage']
    site_index = kwargs['SI_bh_Sb']
    densities = kwargs['densities']
    density_at_bh_age = kwargs['N0_Sb']
    basal_area_at_bh_age = kwargs['BA_Sb0']
    present_basal_area = kwargs['BA_SbT']
    species_proportion_at_bh_age = kwargs['SC_Sb']
    present_density = kwargs['N_bh_SbT']
    factor = 1.2
    factor1 = 1.5 * factor
    tolerance = 0.1
    within_tolerance = False
    iter_count = 0

    while not within_tolerance:
        ba_est = sim_basal_area_sb(age_at_observation, site_index,
                                   density_at_bh_age,
                                   basal_area_at_bh_age,
                                   factor, densities,
                                   fix_proportion_and_density_to_initial_age=True,
                                   species_proportion_at_bh_age=species_proportion_at_bh_age,
                                   present_density=present_density,
                                   stop_at_initial_age=True)[-1]

        if abs(present_basal_area - ba_est) < tolerance:
            within_tolerance = True
        else:
            if (present_basal_area - ba_est) < 0:
                factor1 = factor
                factor_p = factor \
                           * (
                               1 \
                               + (np.log10(present_basal_area) - np.log10(abs(ba_est))) \
                               / (10*np.log10(abs(ba_est))))
                factor = (factor_p+factor)/2
            elif (present_basal_area - ba_est) > 0:
                factor = (factor+factor1)/2

        iter_count += 1

        if iter_count == 1500:
            LOGGER.warning(('Slow convergence with Basal Area: %s'
                            ' and factor:%s '), ba_est, factor)
            break

    return factor


def estimate_basal_area_factor_sw(**kwargs):
    '''White Spruce basal area factor estimator

    This function guarantees that the trajectory of the species basal area
    passes through the basal area measured when the data was collected
    (inventory data)

    :param float startTage: Clock that uses the oldest species as a reference
                            to become the stand age
    :param float SI_bh_Sw: site index of species Sw
    :param float N_bh_SwT: density of sp Sw at time T (it varies over time)
    :param float BA_Sw0: basal area of Sw at breast height age, assumed to be very small
    :param float BA_SwT: basal area of Sw at time T
    :param float N0_Sw: initial density of species Sw at breast height age
    :param float y2bh_Sw: time elapseed in years from zero to breast height age of sp Sw
    :param float startTageSw: species specific ages counted independently
    :param float SC_Sw: proportion of species Sw in the stand
    :param float SDF_Pl0: Stand Density Factor of species Pl
    :param float SDF_Aw0: Stand Density Factor of species Aw
    :param float SDF_Sb0: Stand Density Factor of species Sb
    :param list densities: array of 'densities' objects

    '''
    age_at_observation = kwargs['startTage']
    site_index = kwargs['SI_bh_Sw']
    densities = kwargs['densities']
    density_at_bh_age = kwargs['N0_Sw']
    species_proportion_at_bh_age = kwargs['SC_Sw']
    sdf_aw = kwargs['SDF_Aw0']
    sdf_pl = kwargs['SDF_Pl0']
    sdf_sb = kwargs['SDF_Sb0']
    basal_area_at_bh_age = kwargs['BA_Sw0']
    present_basal_area = kwargs['BA_SwT']
    present_density = kwargs['N_bh_SwT']
    factor = 2.5
    tolerance = 0.1
    within_tolerance = False
    iter_count = 0
    factor1 = 1.5 * factor

    while not within_tolerance:
        ba_est = sim_basal_area_sw(age_at_observation, site_index,
                                   density_at_bh_age, sdf_aw, sdf_pl, sdf_sb,
                                   basal_area_at_bh_age,
                                   factor, densities,
                                   fix_proportion_and_density_to_initial_age=True,
                                   species_proportion_at_bh_age=species_proportion_at_bh_age,
                                   present_density=present_density,
                                   stop_at_initial_age=True)[-1]
        if abs(present_basal_area - ba_est) < tolerance:
            within_tolerance = True
        else:
            if (present_basal_area - ba_est) < 0:
                factor1 = factor
                factor_p = factor \
                           * (1 \
                              + (np.log10(present_basal_area) - np.log10(abs(ba_est))) \
                              / (10*np.log10(abs(ba_est))))
                factor = (factor_p+factor)/2
            elif (present_basal_area - ba_est) > 0:
                factor = (factor + factor1)/2

        iter_count += 1

        if iter_count == 1500:
            LOGGER.warning(('Slow convergence with Basal Area: %s'
                            ' and factor:%s '), ba_est, factor)
            break

    return factor


def estimate_basal_area_factor_pl(**kwargs):
    '''Lodgepole pine basal area factor estimator

    This function guarantees that the trajectory of the species basal area
    passes through the basal area measured when the data was collected
    (inventory data)

    :param float startTage: Clock that uses the oldest species as a reference
                            to become the stand age
    :param float SI_bh_Pl: site index of species Pl
    :param float N_bh_PlT: density of sp Pl at time T (it varies over time)
    :param float BA_Pl0: basal area of Pl at breast height age, assumed to be very small
    :param float BA_PlT: basal area of Pl at time T
    :param float N0_Pl: initial density of species Pl at breast height age
    :param float y2bh_Pl: time elapseed in years from zero to breast height age of sp Pl
    :param float startTagePl: species specific ages counted independently
    :param float SC_Pl: proportion of species Pl in the stand
    :param float SDF_Sw0: Stand Density Factor of species Sw
    :param float SDF_Aw0: Stand Density Factor of species Aw
    :param float SDF_Sb0: Stand Density Factor of species Sb
    :param list densities: array of 'densities' objects

    '''
    age_at_observation = kwargs['startTage']
    site_index = kwargs['SI_bh_Pl']
    density_at_bh_age = kwargs['N0_Pl']
    densities = kwargs['densities']
    sdf_aw = kwargs['SDF_Aw0']
    sdf_sw = kwargs['SDF_Sw0']
    sdf_sb = kwargs['SDF_Sb0']
    basal_area_at_bh_age = kwargs['BA_Pl0']
    present_basal_area = kwargs['BA_PlT']
    species_proportion_at_bh_age = kwargs['SC_Pl']
    present_density = kwargs['N_bh_PlT']
    factor = 100.0
    tolerance = 0.1
    within_tolerance = False
    iter_count = 0
    factor1 = 1.5 * factor

    while not within_tolerance:
        ba_est = sim_basal_area_pl(age_at_observation, site_index,
                                   density_at_bh_age, sdf_aw, sdf_sw, sdf_sb,
                                   basal_area_at_bh_age,
                                   factor, densities,
                                   fix_proportion_and_density_to_initial_age=True,
                                   species_proportion_at_bh_age=species_proportion_at_bh_age,
                                   present_density=present_density,
                                   stop_at_initial_age=True)[-1]

        if abs(present_basal_area - ba_est) < tolerance:
            within_tolerance = True
        else:
            if (present_basal_area - ba_est) < 0:
                factor1 = factor
                factor_p = factor * (present_basal_area / ba_est)
                factor = (factor_p+factor)/2
            elif (present_basal_area - ba_est) > 0:
                factor = (factor + factor1)/2

        iter_count += 1

        if iter_count == 150:
            LOGGER.warning(('Slow convergence with Basal Area: %s'
                            ' and factor:%s '), ba_est, factor)
            break

    return factor



"""Density estimators"""
#pylint: disable=invalid-name,no-member
# The invalid name linting is disabled since this is a math heavy module
# it makes sense to use short names. Please still use sensible names for non
# mathematical variables
# no-member warning is also ignored - probably it is a c extension which
# pylint doesn't see
import numpy as np


def estimate_density_aw(SDF_Aw0, bhage_Aw, SI_bh_Aw, ret_detail=False):
    '''Main purpose of this function is to project densities forward and backward
    in time for the species

    :param float SI_bh_Aw: site index of species Aw
    :param float bhage_Aw: breast height age of species Aw
    :param float SDF_Aw0: Stand Density Factor of species Aw
    :param bool ret_detail: whether additional values should be returned - used
    by other functions to iteratively solve this function for SDF
    '''
    if SDF_Aw0 > 0:
        c0 = 0.717966
        c1 = 6.67468
        b3 = (1+c0) \
             * SDF_Aw0 \
             **((c1+np.log(SDF_Aw0))/SDF_Aw0)
        b2 = (c0/4)*(SDF_Aw0**0.5)**(1/(SDF_Aw0))
        b1 = -((1/((SDF_Aw0/1000.0)**(0.5))) \
            + (np.sqrt(1+np.sqrt(50/(np.sqrt(SDF_Aw0)*np.log(50+1)))))) \
            * np.log(50+1)
        k1 = 1+np.exp(b1 + (b2*SI_bh_Aw) + (b3*np.log(50+1)))
        k2 = 1+np.exp(b1 + (b2*SI_bh_Aw) + (b3*np.log(1+bhage_Aw)))
        N_bh_Aw = SDF_Aw0*k1/k2
    else:
        N_bh_Aw = 0

    if ret_detail:
        return {
            'k1': k1,
            'k2': k2,
            'sdf': SDF_Aw0,
            'density': N_bh_Aw,
        }

    return N_bh_Aw

def estimate_density_sb(SDF_Sb0, tage_Sb, SI_bh_Sb, ret_detail=False):
    '''Main purpose of this function is to project densities forward and backward
    in time for the species

    :param float SI_bh_Sb: site index of species Sb
    :param float tage_Sb: total age of species Sb
    :param float SDF_Sb0: Stand Density Factor of species Sb
    :param float SDF_Aw0: Stand Density Factor of species Aw
    :param bool ret_detail: whether additional values should be returned - used
    by other functions to iteratively solve this function for SDF

    '''
    if SDF_Sb0 > 0:
        c1 = -26.3836
        c2 = 0.166483
        c3 = 2.738569
        b2 = c3
        b3 = c3*(SDF_Sb0**(1/SDF_Sb0))
        b1 = c1/((((SDF_Sb0/1000.0)**0.5)+np.log(50+1))**c2)
        k1 = 1+np.exp(b1+(b2*np.log(SI_bh_Sb))+(b3*np.log(1+50)))
        k2 = 1+np.exp(b1+(b2*np.log(SI_bh_Sb))+(b3*np.log(1+tage_Sb)))
        N_bh_Sb = SDF_Sb0*k1/k2
    else:
        N_bh_Sb = 0

    if ret_detail:
        return {
            'k1': k1,
            'k2': k2,
            'sdf': SDF_Sb0,
            'density': N_bh_Sb,
        }

    return N_bh_Sb


def estimate_density_sw(SDF_Sw0, SDF_Aw0, tage_Sw, SI_bh_Sw, ret_detail=False):
    '''Main purpose of this function is to project densities forward and backward
    in time for the species

    :param float SI_bh_Sw: site index of species Sw
    :param float tage_Sw: total age of species Sw
    :param float SDF_Sw0: Stand Density Factor of species Sw
    :param float SDF_Aw0: Stand Density Factor of species Aw
    :param bool ret_detail: whether additional values should be returned - used
    by other functions to iteratively solve this function for SDF

    '''
    if SDF_Sw0 > 0:
        if SDF_Aw0 == 0:
            z1 = 0
        elif SDF_Aw0 > 0:
            z1 = 1

        c1 = -231.617
        c2 = 1.176995
        c3 = 1.733601
        b3 = c3*(SDF_Sw0**(1/SDF_Sw0))
        b2 = c3
        b1 = (c1/((np.log(SDF_Sw0)+np.log(50+1))**c2))+(z1*((1+(SDF_Aw0/1000.0))**0.5))
        k1 = 1+np.exp(b1+(b2*np.log(SI_bh_Sw))+(b3*np.log(50+1)))
        k2 = 1+np.exp(b1+(b2*np.log(SI_bh_Sw))+(b3*np.log(1+tage_Sw)))
        N_bh_Sw = SDF_Sw0*k1/k2
    else:
        N_bh_Sw = 0

    if ret_detail:
        return {
            'k1': k1,
            'k2': k2,
            'sdf': SDF_Sw0,
            'density': N_bh_Sw,
        }

    return N_bh_Sw


def estimate_density_pl(SDF_Aw0, SDF_Sw0, SDF_Sb0, SDF_Pl0, tage_Pl, SI_bh_Pl, ret_detail=False):
    '''Main purpose of this function is to project densities forward and backward
    in time for the species

    :param float SI_bh_Pl: site index of species Pl
    :param float tage_Pl: total age of species Pl
    :param float SDF_Pl0: Stand Density Factor of species Pl
    :param float SDF_Aw0: Stand Density Factor of species Aw
    :param float SDF_Sb0: Stand Density Factor of species Sb
    :param float SDF_Sw0: Stand Density Factor of species Sw
    :param bool ret_detail: whether additional values should be returned - used
    by other functions to iteratively solve this function for SDF

    '''
    if SDF_Pl0 > 0:
        c1 = -5.25144
        c2 = -483.195
        c3 = 1.138167
        c4 = 1.017479
        c5 = -0.05471
        c6 = 4.11215

        if SDF_Aw0 == 0:
            z1 = 0
        elif SDF_Aw0 > 0:
            z1 = 1

        if SDF_Sw0 == 0:
            z2 = 0
        elif SDF_Sw0 > 0:
            z2 = 1

        if SDF_Sb0 == 0:
            z3 = 0
        elif SDF_Sb0 > 0:
            z3 = 1

        k = (1+(c6*(SDF_Pl0**0.5)))/SDF_Pl0
        b3 = c4*(SDF_Pl0**k)
        b2 = c4/(np.sqrt(SDF_Pl0)**c5)
        b1 = \
             (c1 \
              + (z1*(SDF_Aw0/1000.0)/2.0) \
              + (z2*(SDF_Sw0/1000.0)/3.0) \
              + (z3*(SDF_Sb0/1000.0)/4.0)
             ) \
             + (c2/((SDF_Pl0**0.5)**c3))
        k1 = 1+np.exp(b1 + (b2*np.log(SI_bh_Pl)) + (b3*np.log(50+1)))
        k2 = 1+np.exp(b1 + (b2*np.log(SI_bh_Pl)) + (b3*np.log(1+tage_Pl)))
        N_bh_Pl = SDF_Pl0*k1/k2
    else:
        N_bh_Pl = 0

    if ret_detail:
        return {
            'k1': k1,
            'k2': k2,
            'sdf': SDF_Pl0,
            'density': N_bh_Pl,
        }

    return N_bh_Pl





"""Path utility"""
import os


def _join(*args):
    """Join path elements locally or for s3

    For s3, we need a utility to join paths that will always use '/'. For local
    operations, need the platform specific path separator.

    :param args: path elements
    """
    if args[0].lower() == ('s3://'):
        path = '%s%s' % (
            args[0],
            '/'.join(args[1:]) # what if first element is
            )
    elif args[0].lower().startswith('s3://'):
        path = '/'.join(args)
    else:
        path = os.path.join(*args)

    return path



"""Volume calculations"""
#pylint: disable=invalid-name
# The invalid name linting is disabled since this is a math heavy module
# it makes sense to use short names. Please still use sensible names for non
# mathematical variables
# TODO: Merchantable volume should call the gross total volume functions
from __future__ import division
import numpy as np

DEFAULT_UTILIZATIONS = {
    "aw": {
        "topDiamInsideBark": 7,
        "stumpDiamOutsideBark": 13,
        "stumpHeight": 0.3
    },
    "sw": {
        "topDiamInsideBark": 7,
        "stumpDiamOutsideBark": 13,
        "stumpHeight": 0.3
    },
    "sb": {
        "topDiamInsideBark": 7,
        "stumpDiamOutsideBark": 13,
        "stumpHeight": 0.3
    },
    "pl": {
        "topDiamInsideBark": 7,
        "stumpDiamOutsideBark": 13,
        "stumpHeight": 0.3
    }
}


def _gross_totalvolume_aw(basal_area, top_height):
    ''' White Aspen Gross Total Volume

    Note that inputs may be scalars, or numpy arrays.

    :param float basal_area: basal area
    :param float top_height: top height

    '''
    a1 = 0.248718
    a2 = 0.98568
    a3 = 0.857278
    a4 = -24.9961
    tot_vol = a1 \
              * (basal_area ** a2) \
              * (top_height ** a3) \
              * np.exp(1 \
                       + (a4 / ((top_height ** 2) + 1)))

    return tot_vol


def _gross_totalvolume_sw(basal_area, top_height):
    '''White Spruce Gross Total Volume

    Note that inputs may be scalars, or numpy arrays.

    :param float basal_area: basal area
    :param float top_height: top height

    '''
    b1 = 0.41104
    b2 = 0.983108
    b3 = 0.971061
    tot_vol = b1 \
              * (basal_area ** b2) \
              * (top_height ** b3)

    return tot_vol


def _gross_totalvolume_sb(basal_area, top_height):
    '''Black Spruce Gross Total Volume

    Note that inputs may be scalars, or numpy arrays.

    :param float basal_area: basal area
    :param float top_height: top height

    '''
    b1 = 0.48628
    b2 = 0.982962
    b3 = 0.910603
    tot_vol = b1 \
              * (basal_area ** b2) \
              * (top_height ** b3)

    return tot_vol


def _gross_totalvolume_pl(basal_area, top_height):
    '''Lodgepole Pine Gross Total Volume

    Note that inputs may be scalars, or numpy arrays.

    :param float basal_area: basal area
    :param float top_height: top height

    '''
    a1 = 0.194086
    a2 = 0.988276
    a3 = 0.949346
    a4 = -3.39036
    tot_vol = a1 \
              * (basal_area ** a2) \
              * (top_height ** a3) \
              * np.exp(
                  1 + (a4 / ((top_height ** 2) + 1)))

    return tot_vol

def gross_total_volume(species, *args):
    """Gross total volume

    :param str species: species abbreviation
    :param *args: additional arguments to species specific gross total
                  volume functions

    """
    return {
        'Aw': _gross_totalvolume_aw,
        'Sw': _gross_totalvolume_sw,
        'Sb': _gross_totalvolume_sb,
        'Pl': _gross_totalvolume_pl,
    }[species](*args)


def _merchantable_volume_aw(density, basal_area, top_height, total_volume,
                            stump_dob=13, top_dib=7, stump_height=0.3):
    '''Merchantable volume for white aspen

    Only new variables are the stump diameter outside bark, stump height and
    top diameter inside bark The if below was used (and in other functions) to
    avoid division by zero when density is zero, i.e., when the species is
    absent in the plot.

    :param float density: density
    :param float basal_area: basal area
    :param float top_height: top height
    :param float StumpHeight: stump height
    :param float TopDib: top diameter inside bark
    :param float total_volume: Gross total volume
    :param float StumpDOB:  stump diameter outside bark

    '''
    k = (basal_area * 10000.0 / density) ** 0.5
    b0 = 0.993673
    b1 = 923.5825
    b2 = -3.96171
    b3 = 3.366144
    b4 = 0.316236
    b5 = 0.968953
    b6 = -1.61247
    k1 = total_volume * (k ** b0)
    k2 = (
        b1 \
        * (top_height ** b2) \
        * (stump_dob ** b3) \
        * (stump_height ** b4) \
        * (top_dib ** b5) \
        * (k ** b6)
    ) + k
    merch_vol = k1 / k2

    return merch_vol


def _merchantable_volume_sb(density, basal_area, top_height, total_volume,
                            stump_dob=13, top_dib=7, stump_height=0.3):
    '''Merchantable volume black spruce

    The if below was used (and in other functions) to avoid division by zero
    when density is zero, i.e., when the species is absent in the plot.

    :param float density: density
    :param float top_height: top height
    :param float StumpDOB:  stump diameter outside bark
    :param float StumpHeight: stump height
    :param float TopDib: top diameter inside bark
    :param float total_volume: Gross total volume

    '''
    k = (basal_area * 10000.0 / density) ** 0.5
    b0 = 0.98152
    b1 = 0.678011
    b2 = -1.10256
    b3 = 4.148139
    b4 = 0.511391
    b5 = 1.484988
    b6 = -3.26425
    merch_vol = (total_volume * (k ** b0)) \
                / (
                    (b1 \
                      * (top_height ** b2) \
                      * (stump_dob ** b3) \
                      * (stump_height ** b4) \
                      * (top_dib ** b5) \
                      * (k ** b6)
                    ) + k
                )

    return merch_vol



def _merchantable_volume_sw(density, basal_area, top_height, total_volume,
                            stump_dob=13, top_dib=7, stump_height=0.3):
    '''Merchantable volume for white spruce

    The if below was used (and in other functions) to avoid
    division by zero when density is zero, i.e., when the
    species is absent in the plot.

    :param float density: density
    :param float top_height: top height
    :param float StumpDOB:  stump diameter outside bark
    :param float StumpHeight: stump height
    :param float TopDib: top diameter inside bark
    :param float total_volume: Gross total volume

    '''
    k = (basal_area * 10000.0 / density) ** 0.5
    b0 = 0.996262
    b1 = 7.021736
    b2 = -1.77615
    b3 = 1.91562
    b4 = 0.4111
    b5 = 1.024803
    b6 = -0.80121
    merch_vol = (total_volume * (k ** b0)) \
                / (
                    (b1 \
                     * (top_height ** b2) \
                     * (stump_dob ** b3) \
                     * (stump_height ** b4) \
                     * (top_dib ** b5) \
                     * (k ** b6)
                    ) + k
                )

    return merch_vol



def _merchantable_volume_pl(density, basal_area, top_height, total_volume,
                            stump_dob=13, top_dib=7, stump_height=0.3):
    '''Merchantable volume for lodgepole pine

    The if below was used (and in other functions) to avoid division by zero
    when density is zero, i.e., when the species is absent in the plot.

    :param float density: density
    :param float top_height: top height
    :param float StumpDOB:  stump diameter outside bark
    :param float StumpHeight: stump height
    :param float TopDib: top diameter inside bark
    :param float total_volume: Gross total volume

    '''
    k = (basal_area * 10000.0 / density) ** 0.5
    b0 = 0.989889
    b1 = 1.055091
    b2 = -0.19072
    b3 = 4.915593
    b4 = 0.42574
    b5 = 1.006379
    b6 = -4.87808
    merch_vol = (total_volume * (k ** b0)) \
                / (
                    (b1 \
                     * (top_height ** b2) \
                     * (stump_dob ** b3) \
                     * (stump_height ** b4) \
                     * (top_dib ** b5) \
                     * (k ** b6)
                    ) + k
                )

    return merch_vol


def merchantable_volume(species, *args, **kwargs):
    """Merchantable volume

    :param str species: species abbreviation
    :param *args: additional arguments to species specific gross total
                  volume functions

    """
    return {
        'Aw': _merchantable_volume_aw,
        'Sw': _merchantable_volume_sw,
        'Sb': _merchantable_volume_sb,
        'Pl': _merchantable_volume_pl,
    }[species](*args, **kwargs)



"""Simulation"""
# TODO: change module/funct name - 'forward' is not necessary or strictly true
from __future__ import division
import logging
import datetime
from collections import defaultdict

import pandas as pd

from pygypsy.utils import _log_loop_progress, estimate_species_composition
from pygypsy.basal_area_factor import get_basal_area_factors_for_all_species
from pygypsy.basal_area_simulate import (
    sim_basal_area_aw,
    sim_basal_area_sw,
    sim_basal_area_sb,
    sim_basal_area_pl,
)
from pygypsy.volume import(
    merchantable_volume,
    gross_total_volume,
    DEFAULT_UTILIZATIONS
)
from pygypsy.density import (
    estimate_density_aw,
    estimate_density_pl,
    estimate_density_sb,
    estimate_density_sw,
)
from pygypsy.asaCompileAgeGivenSpSiHt \
    import ComputeGypsyTreeHeightGivenSiteIndexAndTotalAge


LOGGER = logging.getLogger(__name__)
SPECIES = ('Aw', 'Sw', 'Sb', 'Pl')


def _get_initial_basal_area(current_basal_area):
    initial_basal_area = 0.001

    if initial_basal_area > current_basal_area * 0.5:
        initial_basal_area /= 2

    return initial_basal_area


# note: the complexity from basal area/factor finder simulation could be moved
# here. simulation supports using the values from the year of observation for
# all years before the year of observation. that could be enforced here, or
# with an intermediate function. it would be key to get that complexity out of
# the basal area simulation functions to vectorize them
# TODO: this could also be vectorized instead of in a loop
# DESIGN: could use callbacks here instead of calling each of the functions
# DESIGN: should the responsibility to return 0 be in the called functions or
#         in if clauses here?
def simulate_densities_speciescomp_topheight(n_years=250, start_at_data_year=False, **kwargs):
    '''Estimate, species composition, top height for all species along time

    Time is counted independently for each species.

    :param in duration:
    :param float startTage: It uses the oldest species as a reference to
                            become the stand age
    :param float startTageAw, startTageSw, startTageSb, and startTagePl: age
                                                                         for respective
                                                                         species
    :param float SDF_Pl0: Stand Density Factor of species Pl
    :param float SDF_Aw0: Stand Density Factor of species Aw
    :param float SDF_Sb0: Stand Density Factor of species Sb
    :param float SDF_Sw0: Stand Density Factor of species Sw
    :param float SI_bh_Sw: site index of species Sw
    :param float SI_bh_Aw: site index of species Aw
    :param float SI_bh_Sb: site index of species Sb
    :param float SI_bh_Pl: site index of species Pl
    :param float y2bh_Aw: time elapseed in years from zero to breast height age of sp Aw
    :param float y2bh_Sw: time elapseed in years from zero to breast height age of sp Sw
    :param float y2bh_Sb: time elapseed in years from zero to breast height age of sp Sb
    :param float y2bh_Pl: time elapseed in years from zero to breast height age of sp Pl

    '''
    starttage = kwargs['startTage']
    starttageaw = kwargs['startTageAw']
    y2bh_aw = kwargs['y2bh_Aw']
    starttagesw = kwargs['startTageSw']
    y2bh_sw = kwargs['y2bh_Sw']
    starttagesb = kwargs['startTageSb']
    y2bh_sb = kwargs['y2bh_Sb']
    starttagepl = kwargs['startTagePl']
    y2bh_pl = kwargs['y2bh_Pl']
    sdf_aw0 = kwargs['SDF_Aw0']
    sdf_sw0 = kwargs['SDF_Sw0']
    sdf_pl0 = kwargs['SDF_Pl0']
    sdf_sb0 = kwargs['SDF_Sb0']
    si_bh_aw = kwargs['SI_bh_Aw']
    si_bh_sw = kwargs['SI_bh_Sw']
    si_bh_sb = kwargs['SI_bh_Sb']
    si_bh_pl = kwargs['SI_bh_Pl']
    densities_along_time = []

    end_year = n_years if not start_at_data_year else n_years + starttage

    year = 0 if not start_at_data_year else starttage
    tage_aw = starttageaw - starttage if not start_at_data_year else starttageaw
    tage_sw = starttagesw - starttage if not start_at_data_year else starttagesw
    tage_pl = starttagepl - starttage if not start_at_data_year else starttagepl
    tage_sb = starttagesb - starttage if not start_at_data_year else starttagesb

    while year < end_year:
        bhage_aw = tage_aw - y2bh_aw
        bhage_sw = tage_sw - y2bh_sw
        bhage_pl = tage_pl - y2bh_pl
        bhage_sb = tage_sb - y2bh_sb

        if bhage_aw < 0:
            n_bh_awt = 0
        else:
            # bhage instead of tage, as specified by the GYPSY model itself
            n_bh_awt = estimate_density_aw(sdf_aw0, bhage_aw, si_bh_aw)

        if tage_sb < 0:
            n_bh_sbt = 0
        else:
            n_bh_sbt = estimate_density_sb(sdf_sb0, tage_sb, si_bh_sb)

        if tage_sw < 0:
            n_bh_swt = 0
        else:
            n_bh_swt = estimate_density_sw(sdf_sw0, sdf_aw0, tage_sw, si_bh_sw)

        if tage_pl < 0:
            n_bh_plt = 0
        else:
            n_bh_plt = estimate_density_pl(sdf_aw0, sdf_sw0, sdf_sb0, sdf_pl0,
                                           tage_pl, si_bh_pl)

        if n_bh_awt <= 0:
            topheight_aw = 0
        else:
            topheight_aw = ComputeGypsyTreeHeightGivenSiteIndexAndTotalAge(
                'Aw', si_bh_aw, tage_aw
            )

        if n_bh_sbt <= 0:
            topheight_sb = 0
        else:
            topheight_sb = ComputeGypsyTreeHeightGivenSiteIndexAndTotalAge(
                'Sb', si_bh_sb, tage_sb
            )

        if n_bh_swt <= 0:
            topheight_sw = 0
        else:
            topheight_sw = ComputeGypsyTreeHeightGivenSiteIndexAndTotalAge(
                'Sw', si_bh_sw, tage_sw
            )

        if n_bh_plt <= 0:
            topheight_pl = 0
        else:
            topheight_pl = ComputeGypsyTreeHeightGivenSiteIndexAndTotalAge(
                'Pl', si_bh_pl, tage_pl
            )


        sc_aw, sc_sw, sc_sb, sc_pl = estimate_species_composition(
            n_bh_awt, n_bh_sbt,
            n_bh_swt, n_bh_plt
        )

        densities_along_time.append({
            'N_bh_AwT': n_bh_awt, 'N_bh_SwT': n_bh_swt,
            'N_bh_SbT': n_bh_sbt, 'N_bh_PlT': n_bh_plt,
            'SC_Aw': sc_aw, 'SC_Sw': sc_sw,
            'SC_Sb':sc_sb, 'SC_Pl': sc_pl,
            'tage_Aw': tage_aw, 'tage_Sw': tage_sw,
            'tage_Sb': tage_sb, 'tage_Pl': tage_pl,
            'bhage_Aw': bhage_aw, 'bhage_Sw': bhage_sw,
            'bhage_Sb': bhage_sb, 'bhage_Pl': bhage_pl,
            'topHeight_Aw': topheight_aw, 'topHeight_Sw': topheight_sw,
            'topHeight_Sb': topheight_sb, 'topHeight_Pl': topheight_pl
        })
        year += 1
        tage_aw += 1
        tage_sw += 1
        tage_pl += 1
        tage_sb += 1

    return densities_along_time


def _simulate_row(row, utiliz_params=None, n_years=250, backwards=True,
                  year_of_data_acquisition=0):
    """ Simulate a single plot

    :param row: row of pandas data frame
    :param utliz_params: dictionary of utilization parameters
    :param int n_years: number of years to simulate
    :param bool backwards: whether the simulation from year zero to the
                           year of the data should be done
    :param year_of_data_acquisition: year in which data was acquired, used
                                     to set year index

    .. note: access a single row like df.ix[row, ]
    """
    if utiliz_params is None:
        utiliz_params = DEFAULT_UTILIZATIONS

    SI_bh_Aw = row.at['SI_Aw']
    SI_bh_Sw = row.at['SI_Sw']
    SI_bh_Pl = row.at['SI_Pl']
    SI_bh_Sb = row.at['SI_Sb']
    N_bh_AwT = row.at['N_Aw']
    N_bh_SwT = row.at['N_Sw']
    N_bh_PlT = row.at['N_Pl']
    N_bh_SbT = row.at['N_Sb']
    y2bh_Aw = row.at['y2bh_Aw']
    y2bh_Sw = row.at['y2bh_Sw']
    y2bh_Sb = row.at['y2bh_Sb']
    y2bh_Pl = row.at['y2bh_Pl']
    tage_AwT = row.at['tage_Aw']
    tage_SwT = row.at['tage_Sw']
    tage_PlT = row.at['tage_Pl']
    tage_SbT = row.at['tage_Sb']
    BA_AwT = row.at['BA_Aw']
    BA_SwT = row.at['BA_Sw']
    BA_PlT = row.at['BA_Pl']
    BA_SbT = row.at['BA_Sb']
    SDF_Aw0 = row.at['SDF_Aw']
    SDF_Sw0 = row.at['SDF_Sw']
    SDF_Pl0 = row.at['SDF_Pl']
    SDF_Sb0 = row.at['SDF_Sb']
    N0_Aw = row.at['N0_Aw']
    N0_Sw = row.at['N0_Sw']
    N0_Pl = row.at['N0_Pl']
    N0_Sb = row.at['N0_Sb']

    BA_Aw0 = _get_initial_basal_area(BA_AwT)
    BA_Sw0 = _get_initial_basal_area(BA_SwT)
    BA_Sb0 = _get_initial_basal_area(BA_SbT)
    BA_Pl0 = _get_initial_basal_area(BA_PlT)

    SC_Aw, SC_Sw, SC_Sb, SC_Pl = estimate_species_composition(
        N0_Aw, N0_Sb, N0_Sw, N0_Pl
    )

    tageData = [tage_AwT, tage_SwT, tage_PlT, tage_SbT]
    startTageAw = tageData[0]
    startTageSw = tageData[1]
    startTagePl = tageData[2]
    startTageSb = tageData[3]

    tageData = sorted(tageData, reverse=True)
    startTage = int(tageData[0])

    densities = simulate_densities_speciescomp_topheight(
        n_years=n_years,
        start_at_data_year=False if backwards else True,
        startTage=startTage,
        startTageAw=startTageAw,
        y2bh_Aw=y2bh_Aw,
        startTageSw=startTageSw,
        y2bh_Sw=y2bh_Sw,
        startTageSb=startTageSb,
        y2bh_Sb=y2bh_Sb,
        startTagePl=startTagePl,
        y2bh_Pl=y2bh_Pl,
        SDF_Aw0=SDF_Aw0,
        SDF_Sw0=SDF_Sw0,
        SDF_Pl0=SDF_Pl0,
        SDF_Sb0=SDF_Sb0,
        SI_bh_Aw=SI_bh_Aw,
        SI_bh_Sw=SI_bh_Sw,
        SI_bh_Sb=SI_bh_Sb,
        SI_bh_Pl=SI_bh_Pl
    )

    species_factors = defaultdict(lambda: 1) if not backwards else \
                      get_basal_area_factors_for_all_species(
                          startTage=startTage,
                          startTageAw=startTageAw, startTageSb=startTageSb,
                          startTageSw=startTageSw, startTagePl=startTagePl,
                          y2bh_Aw=y2bh_Aw, y2bh_Sb=y2bh_Sb,
                          y2bh_Sw=y2bh_Sw, y2bh_Pl=y2bh_Pl,
                          SC_Aw=SC_Aw, SC_Sb=SC_Sb, SC_Sw=SC_Sw, SC_Pl=SC_Pl,
                          SI_bh_Aw=SI_bh_Aw, SI_bh_Sb=SI_bh_Sb,
                          SI_bh_Sw=SI_bh_Sw, SI_bh_Pl=SI_bh_Pl,
                          N_bh_AwT=N_bh_AwT, N_bh_SbT=N_bh_SbT,
                          N_bh_SwT=N_bh_SwT, N_bh_PlT=N_bh_PlT,
                          N0_Aw=N0_Aw, N0_Sb=N0_Sb, N0_Sw=N0_Sw, N0_Pl=N0_Pl,
                          BA_Aw0=BA_Aw0, BA_Sb0=BA_Sb0,
                          BA_Sw0=BA_Sw0, BA_Pl0=BA_Pl0,
                          BA_AwT=BA_AwT, BA_SbT=BA_SbT,
                          BA_SwT=BA_SwT, BA_PlT=BA_PlT,
                          SDF_Pl0=SDF_Pl0, SDF_Sb0=SDF_Sb0,
                          SDF_Aw0=SDF_Aw0, SDF_Sw0=SDF_Sw0,
                          densities=densities
                      )

    # use_correction_factor_future here is True because aspen was peculiar,
    # empirically demonstrated that using the factor for the whole
    # simulation yielded better results, probably because of variability in
    # density and basal area unique to aspen
    basal_area_aw_arr = sim_basal_area_aw(
        startTage, SI_bh_Aw, N0_Aw, BA_Aw0, SDF_Aw0,
        species_factors['f_Aw'], densities,
        use_correction_factor_future=True if backwards else False,
        stop_at_initial_age=False,
        force_use_densities=False if backwards else True
    )
    basal_area_sb_arr = sim_basal_area_sb(
        startTage, SI_bh_Sb, N0_Sb, BA_Sb0,
        species_factors['f_Sb'], densities,
        use_correction_factor_future=False, stop_at_initial_age=False,
        fix_proportion_and_density_to_initial_age=False,
        species_proportion_at_bh_age=SC_Sb, present_density=N_bh_SbT,
        force_use_densities=False if backwards else True
    )
    basal_area_sw_arr = sim_basal_area_sw(
        startTage, SI_bh_Sw, N0_Sw, SDF_Aw0, SDF_Pl0, SDF_Sb0, BA_Sw0,
        species_factors['f_Sw'], densities,
        use_correction_factor_future=False, stop_at_initial_age=False,
        fix_proportion_and_density_to_initial_age=False,
        species_proportion_at_bh_age=SC_Sw, present_density=N_bh_SwT,
        force_use_densities=False if backwards else True
    )
    basal_area_pl_arr = sim_basal_area_pl(
        startTage, SI_bh_Pl, N0_Pl, SDF_Aw0, SDF_Sw0, SDF_Sb0, BA_Pl0,
        species_factors['f_Pl'], densities,
        use_correction_factor_future=False, stop_at_initial_age=False,
        fix_proportion_and_density_to_initial_age=False,
        species_proportion_at_bh_age=SC_Pl, present_density=N_bh_PlT,
        force_use_densities=False if backwards else True
    )

    output_df_aw = pd.DataFrame(basal_area_aw_arr, columns=['BA_Aw'])
    output_df_sw = pd.DataFrame(basal_area_sw_arr, columns=['BA_Sw'])
    output_df_sb = pd.DataFrame(basal_area_sb_arr, columns=['BA_Sb'])
    output_df_pl = pd.DataFrame(basal_area_pl_arr, columns=['BA_Pl'])

    densities_df = pd.DataFrame(densities)
    output_df = pd.concat(
        [densities_df, output_df_aw, output_df_sw, output_df_sb, output_df_pl],
        axis=1
    )

    for spec in SPECIES:
        output_df['Gross_Total_Volume_%s' % spec] = gross_total_volume(
            spec,
            output_df['BA_%s' % spec],
            output_df['topHeight_%s' % spec]
        )

        output_df['MerchantableVolume%s' % spec] = merchantable_volume(
            spec,
            output_df['N_bh_%sT' % spec],
            output_df['BA_%s' % spec],
            output_df['topHeight_%s' % spec],
            output_df['Gross_Total_Volume_%s' % spec],
            top_dib=utiliz_params[spec.lower()]['topDiamInsideBark'],
            stump_dob=utiliz_params[spec.lower()]['stumpDiamOutsideBark'],
            stump_height=utiliz_params[spec.lower()]['stumpHeight']
        )

    output_df['Gross_Total_Volume_Con'] = output_df['Gross_Total_Volume_Sw'] \
                                          + output_df['Gross_Total_Volume_Sb'] \
                                          + output_df['Gross_Total_Volume_Pl']
    output_df['Gross_Total_Volume_Dec'] = output_df['Gross_Total_Volume_Aw']
    output_df['Gross_Total_Volume_Tot'] = output_df['Gross_Total_Volume_Con'] \
                                          + output_df['Gross_Total_Volume_Dec']
    output_df['MerchantableVolume_Con'] = output_df['MerchantableVolumeSw'] \
                                          + output_df['MerchantableVolumeSb'] \
                                          + output_df['MerchantableVolumePl']
    output_df['MerchantableVolume_Dec'] = output_df['MerchantableVolumeAw']
    output_df['MerchantableVolume_Tot'] = output_df['MerchantableVolume_Con'] \
                                          + output_df['MerchantableVolume_Dec']

    output_df.reset_index(inplace=True)
    output_df = output_df.rename(columns={'index': 'year'})
    output_df['year'] = output_df['year'] + year_of_data_acquisition + 1
    output_df['id_l1'] = int(row['id_l1'])

    return output_df


def simulate_forwards_df(plot_df, **kwargs):
    """Simulate the evolution of plot characteristics through time

    This begins with the simulation of densities, species, and top height.

    With the time series of those items in hand, the simulation from time zero
    (primary succession) to the time of observation can be skipped. If it is
    not skipped, an optimization routine is used to find factors for each
    species. The factors are multiplied with a parameter of the basal area
    increment to ensure that the simulation time timer zero to the time of
    observation passes through a basal area of 0 and the observed basal area.
    Between time zero and the time of observation, fixed values may optionally
    be used for density, species compositon. This optimization is expensive.

    Once the simulation reaches the year of observation, the simulated
    densities, species composition, and top heights are used in the increment
    functions, instead of using fixed values.

    :param plot_df: pandas.DataFrame with plot data
    :param kwargs: keyword arguments to func:`_simulate_row`

    :return: multi-index data frame with plot id and year as index and values
    of the simulated plot characteristics

    .. warning:: the backwards parameter must be set to True in order to
                 utilize the correction factors, which ensure the data passes
                 through 0 and the observation. without the backwards simulation,
                 this is already ensured. However, for Aspen, the factor found
                 is also used for the forward simulation

    """
    output_dict = {}
    n_rows = plot_df.shape[0]

    for _, row in plot_df.iterrows():
        start = datetime.datetime.now()
        _log_loop_progress(_, n_rows)
        plot_id = str(int(row['id_l1']))
        LOGGER.debug('Starting simulation for plot: %s', plot_id)

        try:
            output_df = _simulate_row(row, **kwargs)
            output_dict[plot_id] = output_df
        except Exception as err: #pylint: disable=broad-except
            LOGGER.critical('Unhandled error: %s', err.message)

        end = datetime.datetime.now()
        duration = end - start
        LOGGER.debug('plot %s took %f seconds', plot_id,
                     duration.total_seconds())

    complete_output_df = pd.concat(output_dict.values(), copy=False)
    complete_output_df.set_index(['id_l1', 'year'], inplace=True)

    return complete_output_df



 #pylint: disable=protected-access
"""Command Line Interface"""
# TODO: handling s3/local is clumsy/repetitive in this module
# should factor it out
import os
import logging
from glob import glob

import boto3
import click
import pandas as pd

from pygypsy.scripts import DEFAULT_CONF_FILE
from pygypsy.scripts.callbacks import _load_and_validate_config
from pygypsy.plot import save_plot
from pygypsy.utils import (
    _log_loop_progress,
    _append_file,
    _parse_s3_url,
    _copy_file,
)
from pygypsy.io import df_to_s3_bucket
from pygypsy.data_prep import prep_standtable
from pygypsy.log import setup_logging, CONSOLE_LOGGER_NAME
from pygypsy.forward_simulation import simulate_forwards_df
from pygypsy._version import get_versions
import pygypsy.path as gyppath


LOGGER = logging.getLogger(CONSOLE_LOGGER_NAME)
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
LOG_FILE_NAME = 'pygypsy.log'


@click.group(context_settings=CONTEXT_SETTINGS)
@click.decorators.version_option(version=get_versions()['version'])
@click.option('--verbose', '-v', is_flag=True)
@click.option('--version', is_flag=True)
@click.option('--output-dir', '-o', type=click.Path(exists=False),
              default='pygypsy-output')
@click.pass_context
def cli(ctx, verbose, output_dir, version):  #pylint: disable=unused-argument
    """Growth and Yield Projection System

    Note: 'prep' subcommand must be run before 'simulate'

    """
    setup_logging()
    if verbose:
        LOGGER.setLevel(logging.DEBUG)
        for handler in LOGGER.handlers:
            handler.setLevel(logging.DEBUG)
    LOGGER.debug('Starting pygypsy')

    s3_params = _parse_s3_url(output_dir)
    bucket_name = s3_params['bucket']
    key_prefix = s3_params['prefix']
    bucket_conn = None

    if bucket_name and key_prefix:
        s3 = boto3.resource('s3')
        bucket_conn = s3.Bucket(bucket_name)
    elif bucket_name is None and key_prefix is None:
        if not os.path.isdir(output_dir):
            LOGGER.info('Output directory %s does not exist. Creating it.',
                        output_dir)
            os.mkdir(output_dir)
    else:
        click.Abort('s3 output-dir: %s, is missing a prefix')

    output_dir = key_prefix if key_prefix else output_dir

    ctx.obj = {
        'output-dir': output_dir,
        's3-bucket-name': bucket_name,
        's3-bucket-conn': bucket_conn,
    }


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.pass_context
def generate_config(ctx):
    """Generate a configuration file

    Generates a configuration file, pygypsy-config.json, under DEST directory.

    """
    try:
        output_dir = ctx.obj['output-dir']
        bucket_name = ctx.obj['s3-bucket-name']
        bucket_conn = ctx.obj['s3-bucket-conn']
        output_path = gyppath._join(output_dir, 'pygypsy-config.json')
        _copy_file(DEFAULT_CONF_FILE, output_path, bucket_conn=bucket_conn)
        LOGGER.info('Config file saved at %s', output_path)
    except:
        _copy_file(LOG_FILE_NAME,
                   os.path.join(output_dir, 'generate-config.log'),
                   bucket_conn)
        raise

    _copy_file(LOG_FILE_NAME,
               os.path.join(output_dir, 'generate-config.log'),
               bucket_conn)


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument('standtable', type=click.Path(exists=False))
@click.option('--config-file', '-c', type=click.Path(exists=False),
              default=DEFAULT_CONF_FILE,
              callback=_load_and_validate_config)
@click.pass_context
def prep(ctx, standtable, config_file):
    """Prepare stand data for use in pygpysy simulation"""
    LOGGER.info('Running prep...')
    bucket_name = ctx.obj['s3-bucket-name']
    bucket_conn = ctx.obj['s3-bucket-conn']
    output_dir = ctx.obj['output-dir']
    output_path = gyppath._join(output_dir, 'plot_table_prepped.csv')
    index_label = 'id_l1'

    try:
        standtable_df = pd.read_csv(standtable)
        prepped_data = prep_standtable(
            standtable_df,
            minimum_age=config_file['prep']['minimumAge']
        )

        LOGGER.info('Writing prepped data to %s', output_path)
        if bucket_name:
            df_to_s3_bucket(prepped_data, bucket_conn, output_path,
                            index_label=index_label)
        else:
            prepped_data.to_csv(output_path, index_label=index_label)

    except:
        _copy_file(LOG_FILE_NAME,
                   gyppath._join(output_dir, 'prep.log'),
                   bucket_conn)
        raise

    _copy_file(LOG_FILE_NAME,
               gyppath._join(output_dir, 'prep.log'),
               bucket_conn)


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument('data', type=click.Path(exists=False))
@click.option('--config-file', '-c', type=click.Path(exists=False),
              default=DEFAULT_CONF_FILE,
              callback=_load_and_validate_config)
@click.pass_context
def simulate(ctx, data, config_file):
    """Run pygypsy simulation"""
    bucket_name = ctx.obj['s3-bucket-name']
    bucket_conn = ctx.obj['s3-bucket-conn']
    output_dir = ctx.obj['output-dir']

    try:
        standtable = pd.read_csv(data)
        result = simulate_forwards_df(standtable,
                                      utiliz_params=config_file['utilization'],
                                      n_years=config_file['simulation']['years'],
                                      backwards=config_file['simulation']['backwards'],
                                      year_of_data_acquisition=config_file['data']['acquistionYear'])

        simulation_output_dir = os.path.join(output_dir, 'simulation-data')
        LOGGER.info('Saving output data to %s', simulation_output_dir)

        if bucket_name is None:
            os.mkdir(simulation_output_dir)

        group_level = 1 if config_file['output']['level'] == 'annual' else 0
        for year_or_plot, data in result.groupby(level=group_level):
            filename = '%s.csv' % year_or_plot
            output_path = os.path.join(simulation_output_dir, filename)
            if bucket_name:
                df_to_s3_bucket(data, bucket_conn, output_path)
            else:
                data.to_csv(output_path)
    except:
        _copy_file(LOG_FILE_NAME, gyppath._join(output_dir, 'simulate.log'),
                   bucket_conn)
        raise

    _copy_file(LOG_FILE_NAME, gyppath._join(output_dir, 'simulate.log'),
               bucket_conn)


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument('simulation-output-dir', type=click.Path(exists=True))
@click.option('--config-file', '-c', type=click.Path(exists=False),
              default=DEFAULT_CONF_FILE,
              callback=_load_and_validate_config)
@click.pass_context
def plot(ctx, simulation_output_dir, config_file):
    """Create charts for all files in pygypsy simulation output directory

    """
    output_dir = ctx.obj['output-dir']
    chart_files = glob(os.path.join(simulation_output_dir, '*.csv'))
    LOGGER.info('Plotting all csv files at %s...', simulation_output_dir)

    for i, chart_file in enumerate(chart_files):
        _log_loop_progress(i, len(chart_files))
        chart_df = pd.read_csv(chart_file)
        output_filename = os.path.splitext(os.path.split(chart_file)[-1])[0] + '.png'
        figure_path = os.path.join(simulation_output_dir, 'figures', output_filename)

        save_plot(chart_df, path=figure_path)

    _append_file(LOG_FILE_NAME, os.path.join(output_dir, LOG_FILE_NAME))



""" Helper Functions
"""
from __future__ import division

import os
import errno
import shutil
import logging
from urlparse import urlparse
from copy import deepcopy

from log import CONSOLE_LOGGER_NAME


CONSOLE_LOGGER = logging.getLogger(CONSOLE_LOGGER_NAME)


def _mkdir_p(path):
    """ Make directory recursively
    """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def _log_loop_progress(i, length, n_reports=10):
    step = int(length/n_reports) if length > 10 else 1
    log_steps = range(0, length, step)

    if i in log_steps:
        progress = i/length*100
        CONSOLE_LOGGER.info('%.0f%% complete', progress)


def estimate_species_composition(n_aw, n_sb, n_sw, n_pl):
    '''Calculates species composition based on their densities

    :param float N_Aw, N_Sb, N_Sw, and N_Pl: densities of the respective species

    '''
    n_total = n_aw + n_sb + n_sw + n_pl

    if n_total == 0:
        sc_aw = 0
        sc_sw = 0
        sc_sb = 0
        sc_pl = 0
    else:
        sc_aw = n_aw/n_total
        sc_sw = n_sw/n_total
        sc_sb = n_sb/n_total
        sc_pl = n_pl/n_total

    return sc_aw, sc_sw, sc_sb, sc_pl


def _append_file(source, dest):
    with open(dest, 'wb') as dfd:
        with open(source, 'rb') as sfd:
            shutil.copyfileobj(sfd, dfd)


def _parse_s3_url(url):
    s3_bucket, s3_prefix = None, None

    if url.startswith('s3://'):
        outdir_url = urlparse(url)
        s3_bucket = outdir_url.netloc
        s3_prefix = outdir_url.path.strip('/')

    return {
        'bucket': s3_bucket,
        'prefix': s3_prefix,
    }


def _copy_file_to_s3(bucket_conn, source, dest):
    bucket_conn.upload_file(source, dest)


def _copy_file(source, dest, bucket_conn=None):
    if bucket_conn is None:
        shutil.copyfile(source, dest)
    else:
        _copy_file_to_s3(bucket_conn, source, dest)


def _generate_fplot_dict():
    """Generate 'fplot'

    Given a known dominant species and its site index, and estimation of all
    site indices, generates the 'fplot' dictionary

    Return:
    dict - ???

    ..note: topHeight - top height
    tage - total age
    bhage - breast height age
    N - density
    BA - current Basal Area
    PS - Measured Percent Stocking
    StumpDOB - stump diameter outside bark
    StumpHeight - stump height
    TopDib - top diameter inside bark
    site_index - site index
    PCT - species proportion in plot

    """
    default_species_params = {
        'topHeight': 0, 'tage': 0, 'bhage': 0,
        'N': 0, 'BA': 0, 'PS': 16.9, 'StumpDOB':13,
        'StumpHeight': 0.3, 'TopDib': 7, 'SI': 0, 'PCT': 0
    }
    species = ('Aw', 'Pl', 'Sw', 'Sb')
    fplot_dict = {}

    for s in species:
        fplot_dict[s] = deepcopy(default_species_params)

    return fplot_dict


def _get_gypsy_valid_species(dominant_species):
    """Given the plot dominant species, get the gypsy species

    The dominant species in a plot may not be one of the species in the GYPSY
    model. Douglas Fir for example is not in the GYPSY model, and it is
    appropriate to substitute it with White Spruce.

    :param str dominant_species: abbreviation of plot dominant species
    """
    if dominant_species == 'Pb':
        dominant_species = 'Aw'
    elif dominant_species in ['Fd', 'Fb']:
        dominant_species = 'Sw'

    return dominant_species


def _reclassify_and_sort_species(species_abbrev_perc):
    '''Classify all species in valid gypsy species and sort by percent

    re-classification of species that are not considered in pygypsy as one of
    the species considered in pygypsy (Aw, Sw, Sb, or Pl) and sort the species
    to obtain the dominant species in the plot

    :param list species_abbrev_perc: list of tuples where each element is a
    species abbreviation followed by its percentage

    '''
    species_perc_dict = {'Aw':0, 'Pl':0, 'Sw':0, 'Sb':0}

    for tup in species_abbrev_perc:
        species_abbrev = tup[0]
        species_perc = tup[1]

        if species_abbrev in ['Aw', 'Pb']:
            species_perc_dict['Aw'] = species_perc_dict['Aw'] + species_perc

        elif species_abbrev in ['Sw', 'Fb', 'Fd']:
            species_perc_dict['Sw'] = species_perc_dict['Sw'] + species_perc

        elif species_abbrev == 'Pl':
            species_perc_dict['Pl'] = species_perc_dict['Pl'] + species_perc

        elif species_abbrev == 'Sb':
            species_perc_dict['Sb'] = species_perc_dict['Sb'] + species_perc

    sorted_species_perc_list = [(k, v) for v, k in sorted(
        [(v, k) for k, v in species_perc_dict.items()]
    )]
    sorted_species_perc_list.reverse()

    check_prop1 = sum(species_perc_dict.values())
    if check_prop1 != 100:
        raise ValueError(('Species proportions after grouping '
                          'into 4 species is not correct: %s') % check_prop1)

    return sorted_species_perc_list, species_perc_dict





"""Plotting Convenience Functions"""
import os
import matplotlib.pyplot as plt

from utils import _mkdir_p


def _plot_simulation_variables(simulation_df, axis=None, plot_vars=None, y_lab=''):
    """
    :param simulation_df: output of pygypsy simulation
    :param axis: axes object
    :param plot_vars: list of strings identifying variable (column names) to plot
    :param y_lab: y axis label
    """

    if plot_vars is None:
        raise ValueError('Variable to plot must be specified')

    simulation_vars = simulation_df.loc[:, plot_vars]
    simulation_vars.plot(ax=axis)
    axis.set_xlabel('Year', fontsize=10)
    axis.set_ylabel(y_lab, fontsize=10)
    axis.legend(loc=2, prop={'size':6})
    axis.tick_params(axis='both', which='major', labelsize=8)

def _plot_basal_area(simulation_dataframe, axis=None): #pylint: disable=missing-docstring
    _plot_simulation_variables(simulation_dataframe, axis=axis,
                               plot_vars=['BA_Aw', 'BA_Sw', 'BA_Sb', 'BA_Pl'],
                               y_lab='BA (m2)')

def _plot_merch_volume(simulation_dataframe, axis): #pylint: disable=missing-docstring
    _plot_simulation_variables(simulation_dataframe, axis=axis,
                               plot_vars=['MerchantableVolumeAw', 'MerchantableVolumeSw',
                                          'MerchantableVolumeSb', 'MerchantableVolumePl'],
                               y_lab='Merc. Vo. (m3)')

def _plot_merch_volume_conif_decid(simulation_dataframe, axis): #pylint: disable=missing-docstring
    _plot_simulation_variables(simulation_dataframe, axis=axis,
                               plot_vars=['MerchantableVolume_Con',
                                          'MerchantableVolume_Dec',
                                          'MerchantableVolume_Tot'],
                               y_lab='Merc. Vo. (m3)')

def _plot_top_height(simulation_dataframe, axis=None): #pylint: disable=missing-docstring
    _plot_simulation_variables(simulation_dataframe, axis=axis,
                               plot_vars=['topHeight_Aw', 'topHeight_Sw',
                                          'topHeight_Sb', 'topHeight_Pl'],
                               y_lab='Top Height (m)')


def _plot_gross_total_volume(simulation_dataframe, axis): #pylint: disable=missing-docstring
    _plot_simulation_variables(simulation_dataframe, axis=axis,
                               plot_vars=['Gross_Total_Volume_Aw', 'Gross_Total_Volume_Sw',
                                          'Gross_Total_Volume_Sb', 'Gross_Total_Volume_Pl'],
                               y_lab='Gr. Tot. Vol. (m3)')

def _plot_gross_total_volume_conif_decid(simulation_dataframe, axis): #pylint: disable=missing-docstring
    _plot_simulation_variables(simulation_dataframe, axis=axis,
                               plot_vars=['Gross_Total_Volume_Con',
                                          'Gross_Total_Volume_Dec',
                                          'Gross_Total_Volume_Tot'],
                               y_lab='Gr. Tot. Vol. (m3)')


def _plot_species_composition(simulation_dataframe, axis): #pylint: disable=missing-docstring
    _plot_simulation_variables(simulation_dataframe, axis=axis,
                               plot_vars=['SC_Aw', 'SC_Sw', 'SC_Sb', 'SC_Pl'],
                               y_lab='Sp. Comp.')

def _plot_density(simulation_dataframe, axis): #pylint: disable=missing-docstring
    _plot_simulation_variables(simulation_dataframe, axis=axis,
                               plot_vars=['N_bh_AwT', 'N_bh_SwT', 'N_bh_SbT', 'N_bh_PlT'],
                               y_lab='Density')

def save_plot(simulation_dataframe, path):
    '''Save plots of pygypsy simulation output

    Creates a panel and includes all plots generated as pygypsy outputs
    (simulation_dataframe) and saves the panel on a folder determined byt path

    '''
    _mkdir_p(os.path.dirname(path))
    fig = plt.figure(1)
    sub1 = fig.add_subplot(321)
    sub2 = fig.add_subplot(322)
    sub3 = fig.add_subplot(323)
    sub4 = fig.add_subplot(324)
    sub5 = fig.add_subplot(325)
    sub6 = fig.add_subplot(326)

    _plot_basal_area(simulation_dataframe, axis=sub1)
    _plot_merch_volume_conif_decid(simulation_dataframe, axis=sub2)
    _plot_top_height(simulation_dataframe, axis=sub3)
    _plot_gross_total_volume_conif_decid(simulation_dataframe, axis=sub4)
    _plot_species_composition(simulation_dataframe, axis=sub5)
    _plot_density(simulation_dataframe, axis=sub6)

    plt.tight_layout()
    plt.savefig(path) #TODO: specify page size here to reduce legend size
    plt.close()

    return True



"""Increment functions

This module contains functions to increment plot characteristics at an annual
resolution.

It uses the non-spatial implementaion.

"""
# NOTE: cython has a bug with ** and negative integer exponents, it coerces result to int
# pow(base, exponent) should be used if it is desired to use a negative integer with float
import numpy as np
cimport numpy as np

def increment_basal_area_aw(np.float64_t SC,
                            np.float64_t SI_bh,
                            np.float64_t N_bh,
                            np.float64_t N0,
                            np.float64_t bhage,
                            np.float64_t BA):
    '''Predicts the increment in basal area for the subsequent year for the
    species.

    This function is also used to make basal area projections and as a function
    called by factor finder functions during estimation of correction factor
    for the species.

    :param float SI_bh: site index of species Aw
    :param float bhage: breast height age of speceis Aw
    :param float N_bh: density of species Aw
    :param float SC: proportion of species Aw in the stand
    :param float BA: Basal area of the species Aw
    :param float N0: initial density of species Aw at breast height age

    '''
    cdef np.float64_t BAinc, a1, a2, a3, a4, a5, X1, X2, X3, X4, X5, X6, d1, d2, \
        k, n, d
    if N_bh == 0:
        BAinc = 0

    if bhage < 0:
        bhage = 0

    if BA < 0:
        BAinc = 0
    elif N_bh > 0 and SI_bh > 0:
        a1 = 0.751313
        a2 = 0.018847
        a3 = 1.143762
        a4 = -0.03475
        a5 = 0.835189
        X1 = a1 * pow(10, -4)
        X2 = bhage ** 2
        X3 = np.exp(-a2 * bhage ** (0.5 + a1))
        X4 = SC ** a5
        X5 = (np.log(1 + N0 * (1 + bhage) ** 0.5)) ** 2
        X6 = SI_bh
        d1 = (1 + BA) ** a3
        d2 = 1 + np.exp(1 -np.log(1 + SC ** 2) / 2.0)
        k = a4 * np.log(0.01 + bhage / 10.0)
        n = X1 * X2 * X3 * X4 * X5 * X6
        d = d1 * d2
        BAinc = (n / d) + k

    return  BAinc


def increment_basal_area_sb(np.float64_t SC,
                            np.float64_t SI_bh,
                            np.float64_t N_bh,
                            np.float64_t N0,
                            np.float64_t bhage,
                            np.float64_t BA):
    '''Predicts the increment in basal area for the subsequent year for the
    species.

    This function is used to make basal area projections and as a function
    called by factor finder functions during estimation of correction factor
    for the species.

    :param float SI_bh: site index of species Sb
    :param float bhage: breast height age of speceis Sb
    :param float N_bh: density of species Sb
    :param float SC: proportion of species Sb in the stand
    :param float BA: Basal area of the species Sb
    :param float N0: initial density of species Sb at total age

    '''
    cdef np.float64_t BAinc, a1, a2, a3, k, k1

    if N_bh == 0:
        BAinc = 0

    if bhage < 0:
        bhage = 0

    if BA < 0:
        BA = 0

    elif N_bh > 0 and SI_bh > 0:
        a1 = 0.966285
        a2 = 0.056315
        a3 = 0.17191

        k = ( 1 + ((N0 ** 0.5) * ((1 + bhage) ** 0.5))) \
            * ( np.exp(-(N0 / 4.0) / 10000.0)) \
            * (np.log(1 + SI_bh)) \
            / ((1 + BA) ** a2)

        k1 = (pow(10, -4)) \
             * a1 \
             * (np.exp(-a2 * bhage)) \
             * (SC ** a3) \
             * (bhage ** (a2 + np.sqrt(a1)))

        BAinc = k * k1

    return BAinc


def increment_basal_area_sw(np.float64_t SC,
                            np.float64_t SI_bh,
                            np.float64_t N_bh,
                            np.float64_t N0,
                            np.float64_t bhage,
                            np.float64_t SDF_Aw0,
                            np.float64_t SDF_Pl0,
                            np.float64_t SDF_Sb0,
                            np.float64_t BA):
    '''Predicts the increment in basal area for the subsequent year for the
    species.

    This function is used to make basal area projections and as a function
    called by factor finder functions during estimation of correction factor
    for the species.

    :param float SI_bh: site index of species Sw
    :param float bhage: breast height age of speceis Sw
    :param float N_bh: density of species Sw
    :param float SC: proportion of species Sw in the stand
    :param float BA: Basal area of the species Sw
    :param float N0: initial density of species Sw at total age
    :param float SDF_Pl0: Stand Density Factor of species Pl
    :param float SDF_Aw0: Stand Density Factor of species Aw
    :param float SDF_Sb0: Stand Density Factor of species Sb

    '''
    cdef np.float64_t BAinc, \
        a1, a2, a3, a4, a5, a6, \
        z1, z2, z3, k, k1, k2, m

    if N_bh == 0:
        BAinc = 0

    if bhage < 0:
        bhage = 0

    if BA < 0:
        BA = 0

    if N_bh > 0 and SI_bh > 0:
        a1 = 0.089153
        a2 = 0.072171
        a3 = -0.11483
        a4 = 5.839408
        a5 = 1.753002
        a6 = 0.239521

        if SDF_Aw0 == 0:
            z1 = 0
        elif SDF_Aw0 > 0:
            z1 = 1
        if SDF_Pl0 == 0:
            z2 = 0
        elif SDF_Pl0 > 0:
            z2 = 1
        if SDF_Sb0 == 0:
            z3 = 0
        elif SDF_Sb0 > 0:
            z3 = 1

        k = (a4 * z1 * np.log(1 + (SDF_Aw0 / 10000.0 ))) \
            + (a5 * z2 * np.log(1 + (SDF_Pl0 / 10000.0))) \
            + (z3 * np.log(1 + (SDF_Sb0 / 10000.0)))

        k1 = (pow(10, -4)) \
             * a1 \
             * ((a2 + bhage) ** 2) \
             * ((1 + bhage) ** ((a1 ** 0.5) + a2 - a3)) \
             * np.exp(-a2 * bhage)\
             * (SC ** a6)

        k2 = 1 \
             + np.exp(
                 1 \
                 + k \
                 + ((np.log(1 + ((N0 ** 0.5) / 10000.0))) / 2.0) \
                 + (a3 * np.log(1 + BA))
             )

        m = (np.log(1 + (N0 * ((1 + bhage) ** 0.5))) ** 2) \
            * (SI_bh ** 0.5) \
            * np.exp(-(N0 / 10.0) / 10000.0)

        BAinc = k1 * m / k2

    return BAinc


def increment_basal_area_pl(np.float64_t SC,
                            np.float64_t SI_bh,
                            np.float64_t N_bh,
                            np.float64_t N0,
                            np.float64_t bhage,
                            np.float64_t SDF_Aw0,
                            np.float64_t SDF_Sw0,
                            np.float64_t SDF_Sb0,
                            np.float64_t BA):
    '''Predicts the increment in basal area for the subsequent year for the
    species.

    This function is used to make basal area projections and as a
    function called by factor finder functions during estimation of correction
    factor for the species.

    :param float SC: proportion of species Pl in the stand
    :param float SI_bh: site index of species Pl
    :param float bhage: breast height age of species Pl
    :param float N_bh: density of species Pl
    :param float BA: Basal area of the species Pl
    :param float N0: initial density of species Pl at total age
    :param float SDFSw0: Stand Density Factor of species Sw
    :param float SDF_Aw0: Stand Density Factor of species Aw
    :param float SDF_Sb0: Stand Density Factor of species Sb

    '''
    cdef np.float64_t BAinc, a1, a2, a3, a4, a5, a6, \
        z1, z2, z3, k, k1, k2, m1, m2

    if N_bh == 0:
        BAinc = 0

    if bhage < 0:
        bhage = 0

    if BA <= 0 or SC == 0:
        BA = 0
        BAinc = 0

    if N_bh > 0 and SI_bh > 0:
        a1 = 3.923984
        a2 = 0.05752
        a3 = 0.560402
        a4 = 0.672506
        a5 = -0.00358
        a6 = 0.775765

        if SDF_Aw0 == 0:
            z1 = 0
        elif SDF_Aw0 > 0:
            z1 = 1
        if SDF_Sw0 == 0:
            z2 = 0
        elif SDF_Sw0 > 0:
            z2 = 1
        if SDF_Sb0 == 0:
            z3 = 0
        elif SDF_Sb0 > 0:
            z3 = 1

        k = (z1 * np.log(1 + (SDF_Aw0 / 1000.0))) \
            + (z2 * (np.log(1 + (SDF_Sw0 / 1000.0))) / 2.0)\
            + (z3 * (np.log(1 + (SDF_Sb0 / 1000.0))) / 2.0)

        k1 = pow(10, -4) \
             * a1 \
             * bhage \
             * np.exp(-a2* bhage) \
             * (1 + ((np.log(1 + bhage)) / 2.0))

        k2 = 1 \
             + np.exp(
                 ( k / 2.0) \
                 + np.log(1 + ((N0 / 3.0) / 10000.0)) \
                 - (a3 * (SC ** 0.5)) \
                 + (a4 * np.log(1 + BA))
             )

        m1 = (1 + a3 + (SI_bh ** a6)) \
             * (N0 ** 0.5) \
             * np.exp(-(N0 / 3.0) / 10000.0)

        m2 = a5 * np.log(0.01 + (bhage / 10.0))

        BAinc = (k1 * m1 / k2) + m2

    return BAinc



#pylint: disable=missing-docstring
import os

CONF_DIR = os.path.join(os.path.dirname(__file__), 'config')
DEFAULT_CONF_FILE = os.path.join(CONF_DIR, 'default-conf.json')
CONF_SCHEMA_FILE = os.path.join(CONF_DIR, 'conf.schema')



"""pygypsy

Based on Hueng et all (2009)

Huang, S., Meng, S. X., & Yang, Y. (2009). A growth and yield projection system
(GYPSY) for natural and post-harvest stands in Alberta. Forestry Division,
Alberta Sustainable Resource Development, 25.

Important Acronyms:

aw = white aspen
sb = black spruce
sw = white spruce
pl = logdepole pine
bhage = Breast Height Age
tage = Total age
si_<xx> = estimated site index for species <xx>
y2bh = years until breast height age can be measured

"""
import os
import matplotlib
from ._version import get_versions

import pygypsy.basal_area_increment

__version__ = get_versions()['version']
del get_versions

# Force matplotlib to not use any Xwindows backend so that headless docker works
matplotlib.use('Agg')

__all__ = ['basal_area_increment']




from __future__ import division

import logging
import numpy


LOGGER = logging.getLogger(__name__)


def ComputeGypsySiteIndex(species = 'Aw', totalHeight = 1.3, bhage = 0, totalAge = 0):
    '''
    Total age is unknown
    Breast height age is known
    Total height is known

    Equations and logic derived from:
    Huang, S., Meng, S.X., and Yang, Y. 2009. A growth and yield projection system (GYPSY) for natural and
    post-harvest stands in Alberta. Technical Report Pb. No.:T/216. Forest Management Branch, Alberta Sustainable
    Resource Development, Edmonton, AB, CAN.  Appendix I. Pp. 21,22.

    Error in white spruce equation discovered Dec 5 2014

    '''
    tage = totalAge
    si0 = 0
    iterCount = 0
    if totalHeight > 1.3:
        if bhage > 0 or totalAge > 0:
            if species == 'P' or \
               species == 'Pl' or \
               species == 'Pj' or \
               species == 'Pa' or \
               species == 'Pf':
                b1 = 12.84571
                b2 = -5.73936
                b3 = -0.91312
                b4 = 0.150668
                si0 = 10
                si1 = 0
                while abs(si0-si1)>0.00000001:
                    k1 = numpy.exp(b1+b2*(numpy.log(50+1)**0.5) + b3*numpy.log(si0)+b4*(50**0.5))
                    k2 = si0**b3
                    k3 = ((si0)*(1+k1)/1.3-1)/(numpy.exp(b1)*numpy.exp(b4*(50**0.5))*k2)
                    y2bh = numpy.exp((numpy.log(k3)/b2)**2)-1
                    if bhage > 0 and totalAge == 0:
                        tage = bhage + y2bh
                    else:
                        #bhage = 0 and totalGe > 0
                        #tage = totalAge
                        bhage = tage - y2bh
                    x10 = (1+numpy.exp(b1+b2*(numpy.log(tage+1)**0.5)+b3*numpy.log(si0)+b4*(50**0.5)))
                    x20 = (1+numpy.exp(b1+b2*(numpy.log(50+1)**0.5)+b3*numpy.log(si0)+b4*(50**0.5)))
                    si1 = totalHeight*(x10/x20)
                    si0 = (si0+si1)/2
                SI_t=si1
                # estimating Site index BH
                k11=1+numpy.exp(b1+b2*(numpy.log(50+1)**0.5)+b3*numpy.log(SI_t)**1+b4*(50**0.5))
                k21= 1+numpy.exp(b1+b2*(numpy.log(50+y2bh+1)**0.5)+b3*numpy.log(SI_t)+b4*(50**0.5))
                SI_bh=SI_t*k11/k21


            if species == 'Sw' or \
               species == 'Se' or \
               species == 'Fd' or \
               species == 'Fb' or \
               species == 'Fa':
                b1 = 12.14943
                b2 = -3.77051
                b3 = -0.28534
                b4 = 0.165483
                si0 = 10
                si1 = 0
                while abs(si0-si1)>0.00000001:
                    k1 = numpy.exp(b1+b2*(numpy.log(50**2+1)**0.5) + b3*(numpy.log(si0)**2)+b4*(50**0.5))
                    k2 = si0**(b3*numpy.log(si0))
                    k3 = ((si0)*(1+k1)/1.3-1)/(numpy.exp(b1)*numpy.exp(b4*(50**0.5))*k2)
                    y2bh = (numpy.exp((numpy.log(k3)/b2)**2)-1)**0.5
                    if bhage > 0 and totalAge == 0:
                        tage = bhage + y2bh
                    else:
                        #bhage = 0 and totalGe > 0
                        #tage = totalAge
                        bhage = tage - y2bh
                    #x10 = (1+numpy.exp(b1+b2*(numpy.log((tage**2)+1)**0.5)+b3*(numpy.log(si0)**2)+b4*(50**0.5)))
                    #x20 = (1+numpy.exp(b1+b2*(numpy.log((50**2)+1)**0.5)+b3*(numpy.log(si0)**2)+b4*(50**0.5)))
                    x10 = (1+numpy.exp(b1+b2*((numpy.log((tage**2)+1))**0.5)+b3*(numpy.log(si0)**2)+b4*(50**0.5)))
                    x20 = (1+numpy.exp(b1+b2*((numpy.log((50**2)+1))**0.5)+b3*(numpy.log(si0)**2)+b4*(50**0.5)))
                    si1 = totalHeight*(x10/x20)
                    si0 = (si0+si1)/2
                    iterCount = iterCount + 1
                    if iterCount == 1000:
                        LOGGER.warning(('Site index Routine Terminated; iter > 1000 '
                                        'Sp = %s, treeHt = %s, treeSi = %s, treeAge = %s'),
                                       species, totalHeight, si0, totalAge)
                    SI_bh=0
                SI_t=si1
                # estimating Site index BH
                k11=1+numpy.exp(b1+b2*(((numpy.log(50**2+1)))**0.5)+b3*(numpy.log(SI_t))**2+b4*(50**0.5) )
                k21=1+numpy.exp(b1+b2*(numpy.log(1+(50+y2bh)**2)**0.5)+b3*((numpy.log(SI_t))**2)+b4*(50**0.5) )
                SI_bh=SI_t*k11/k21

            if species == 'Sb' or \
               species == 'Lt' or \
               species == 'La' or \
               species == 'Lw' or \
               species == 'L':
                b1 = 14.56236
                b2 = -6.04705
                b3 = -1.53715
                b4 = 0.240174
                si0 = 10
                si1 = 0
                while abs(si0-si1)>0.00000001:
                    k1 = numpy.exp(b1+b2*(numpy.log(50+1)**0.5) + b3*numpy.log(si0)+b4*(50**0.5))
                    k2 = si0**b3
                    k3 = ((si0)*(1+k1)/1.3-1)/(numpy.exp(b1)*numpy.exp(b4*(50**0.5))*k2)
                    y2bh = numpy.exp((numpy.log(k3)/b2)**2)-1
                    if bhage > 0 and totalAge == 0:
                        tage = bhage + y2bh
                    else:
                        #bhage = 0 and totalGe > 0
                        #tage = totalAge
                        bhage = tage - y2bh
                    x10 = (1+numpy.exp(b1+b2*(numpy.log(tage+1)**0.5)+b3*numpy.log(si0)+b4*(50**0.5)))
                    x20 = (1+numpy.exp(b1+b2*(numpy.log(50+1)**0.5)+b3*numpy.log(si0)+b4*(50**0.5)))
                    si1 = totalHeight*(x10/x20)
                    si0 = (si0+si1)/2
                    SI_bh=0
                SI_t=si1
                # estimating Site index BH
                k11=1+numpy.exp(b1+b2*(numpy.log(50+1)**0.5)+b3*numpy.log(SI_t)+b4*(50**0.5))
                k21=1+numpy.exp(b1+b2*((numpy.log(50+y2bh+1))**0.5)+b3*numpy.log(SI_t)+b4*(50**0.5))
                SI_bh=SI_t*k11/k21

            if species == 'Aw' or \
               species == 'Bw' or \
               species == 'Pb' or \
               species == 'A' or \
               species == 'H':
                b1 = 9.908888
                b2 = -3.92451
                b3 = -0.32778
                b4 = 0.134376
                si0 = 10
                si1 = 0
                while abs(si0-si1)>0.00000001:
                    k1 = numpy.exp(b1+b2*(numpy.log(50+1)**0.5) + b3*(numpy.log(si0)**2)+b4*(50**0.5))
                    k2 = si0**(b3*numpy.log(si0))
                    k3 = ((si0)*(1+k1)/1.3-1)/(numpy.exp(b1)*numpy.exp(b4*(50**0.5))*k2)
                    y2bh = numpy.exp((numpy.log(k3)/b2)**2)-1
                    if bhage > 0 and totalAge == 0:
                        tage = bhage + y2bh
                    else:
                        #bhage = 0 and totalGe > 0
                        tage = totalAge
                        bhage = tage - y2bh
                    x10 = (1+numpy.exp(b1+b2*(numpy.log(tage+1)**0.5)+b3*(numpy.log(si0)**2)+b4*(50**0.5)))
                    x20 = (1+numpy.exp(b1+b2*(numpy.log(50+1)**0.5)+b3*(numpy.log(si0)**2)+b4*(50**0.5)))
                    si1 = totalHeight*(x10/x20)
                    si0 = (si0+si1)/2
                    SI_bh=0
                SI_t=si1
                # estimating Site index BH
                SI_bh=SI_t*(1+numpy.exp(b1+b2*(numpy.log(50+1)**0.5)+b3*(numpy.log(SI_t))**2+b4*(50**0.5)))/(1+numpy.exp(b1+b2*(numpy.log(50+y2bh+1)**0.5)+b3*((numpy.log(SI_t))**2)+b4*(50**0.5)))
        else:
            SI_t = 0
            SI_bh = 0
            y2bh = None
    else:
        if totalHeight<1.3:
            startHt = 0.01
            a = numpy.log(totalHeight/startHt)/totalAge
            y2bh=numpy.log(1.3/startHt)/a
        else:
            y2bh = totalAge


        guessSI = 10 # best SI_t guess
        tolerance = 0.00001
        within_tolerance = False
        iter_count = 0

        while within_tolerance == False:
            est_y2bh = computeGypsyY2BHGivenSpSI(species, guessSI)

            if abs(est_y2bh-y2bh) < tolerance:
                within_tolerance = True
                SI_t = guessSI
            else:
                guessSI = (((est_y2bh+y2bh)/2)/y2bh)*guessSI
                iter_count += 1
                if iter_count == 1000:
                    LOGGER.warning('Slow convergence')
                    break
        SI_bh = SI_t
        tage = totalAge
        bhage = tage - y2bh

    return bhage, tage, SI_t, y2bh, SI_bh

def ComputeGypsyTreeHeightGivenSiteIndexAndTotalAge(species = 'Aw', siteIndex = 1.3, totalAge = 0):
    '''
    Equations and logic derived from:
    Huang, S., Meng, S.X., and Yang, Y. 2009. A growth and yield projection system (GYPSY) for natural and
    post-harvest stands in Alberta. Technical Report Pb. No.:T/216. Forest Management Branch, Alberta Sustainable
    Resource Development, Edmonton, AB, CAN.  Appendix I. Pp. 21,22.

    Error in white spruce equation discovered Dec 5 2014

    '''
    tage = totalAge
    si0 = siteIndex
    treeHeight = 0
    if siteIndex > 1.3:
        if totalAge > 0:
            if species == 'P' or \
               species == 'Pl' or \
               species == 'Pj' or \
               species == 'Pa' or \
               species == 'Pf':
                b1 = 12.84571
                b2 = -5.73936
                b3 = -0.91312
                b4 = 0.150668
                x10 = (1+numpy.exp(b1+b2*(numpy.log(50+1)**0.5)+b3*numpy.log(si0)+b4*(50**0.5)))
                x20 = (1+numpy.exp(b1+b2*(numpy.log(totalAge+1)**0.5)+b3*numpy.log(si0)+b4*(50**0.5)))
                treeHeight = siteIndex*(x10/x20)
            if species == 'Sw' or \
               species == 'Se' or \
               species == 'Fd' or \
               species == 'Fb' or \
               species == 'Fa':
                b1 = 12.14943
                b2 = -3.77051
                b3 = -0.28534
                b4 = 0.165483
                #x10 = (1+numpy.exp(b1+b2*(numpy.log((50**2)+1)**0.5)+b3*((numpy.log(si0))**2)+b4*(50**0.5)))
                #x20 = (1+numpy.exp(b1+b2*(numpy.log((totalAge**2)+1)**0.5)+b3*((numpy.log(si0))**2)+b4*(50**0.5)))
                #treeHeight = siteIndex*(x10/x20)
                x10 = (1+numpy.exp(b1+b2*((numpy.log((50**2)+1))**0.5)+b3*((numpy.log(si0))**2)+b4*(50**0.5)))
                x20 = (1+numpy.exp(b1+b2*((numpy.log((totalAge**2)+1))**0.5) +b3*((numpy.log(si0))**2)+b4*(50**0.5)))
                treeHeight = siteIndex*(x10/x20)

            if species == 'Sb' or \
               species == 'Lt' or \
               species == 'La' or \
               species == 'Lw' or \
               species == 'L':
                b1 = 14.56236
                b2 = -6.04705
                b3 = -1.53715
                b4 = 0.240174
                x10 = (1+numpy.exp(b1+b2*(numpy.log(50+1)**0.5)+b3*numpy.log(si0)+b4*(50**0.5)))
                x20 = (1+numpy.exp(b1+b2*(numpy.log(totalAge+1)**0.5)+b3*numpy.log(si0)+b4*(50**0.5)))
                treeHeight = siteIndex*(x10/x20)
            if species == 'Aw' or \
               species == 'Bw' or \
               species == 'Pb' or \
               species == 'A' or \
               species == 'H':
                b1 = 9.908888
                b2 = -3.92451
                b3 = -0.32778
                b4 = 0.134376
                x10 = (1+numpy.exp(b1+b2*(numpy.log(50+1)**0.5)+b3*(numpy.log(si0)**2)+b4*(50**0.5)))
                x20 = (1+numpy.exp(b1+b2*(numpy.log(totalAge+1)**0.5)+b3*(numpy.log(si0)**2)+b4*(50**0.5)))
                treeHeight = siteIndex*(x10/x20)
        else:
            treeHeight = 0
    else:
            treeHeight = 0

    return treeHeight

def computeTreeAge(siSp='',treeHt = 20, treeSi=15, maxTreeAge = 450,
                   rowIndex = 0):
    if treeHt >1.3:
        maxTreeAgeAw = 200
        maxTreeAgePl = 250
        maxTreeAgeSw = 450
        maxTreeAgeSb = 450
        maxTreeAge = {'Aw': maxTreeAgeAw, 'Pl': maxTreeAgePl, 'Sw': maxTreeAgeSw, 'Sb': maxTreeAgeSb}
        treeAge = (treeHt/treeSi)*25
        htDiffFlag = False
        iterCount = 0
        acceptableDiff = 0.0001
        while htDiffFlag == False:
            newHt = ComputeGypsyTreeHeightGivenSiteIndexAndTotalAge(siSp, treeSi, treeAge)
            if abs(treeHt-newHt) < acceptableDiff:
                htDiffFlag = True
            else:
                #treeAge = ((treeHt-newHt)/treeHt)*(treeAge/2) + treeAge
                treeAge = ((treeHt-newHt)/treeHt)*(treeAge) + treeAge
            iterCount = iterCount + 1

            if iterCount == 150:
                LOGGER.info(('Slow convergence'
                             'Sp = %s, treeHt = %s, treeSi = %s, treeAge = %s'),
                            siSp, treeHt, treeSi, treeAge)
            if treeAge > 1000:
                htDiffFlag = True
                LOGGER.warning(('Tree Age Search Routine Terminated; treeAge > 1000 '
                                'Sp = %s, treeHt = %s, treeSi = %s, treeAge = %s'),
                               siSp, treeHt, treeSi, treeAge)
                treeAge = maxTreeAge[siSp]
            if treeAge < 0:
                htDiffFlag = True
                LOGGER.warning('Slow convergence with negative treeAge: %s', treeAge)
                treeAge = 0
    else:
        y2bh = computeGypsyY2BHGivenSpSI(siSp, treeSi)
        ht = 1.3
        startHt = 0.01 # start tree height assumed tom be 1 cm

        a = numpy.log(ht/startHt)/y2bh
        treeAge = numpy.log(treeHt/startHt)/a

    return treeAge



def computeGypsyY2BHGivenSpSI(species = 'Aw', si0 = 5):
    '''
    Compute years to breast height given species and site index

    Huang, S., Meng, S.X., and Yang, Y. 2009. A growth and yield projection system (GYPSY) for natural and
    post-harvest stands in Alberta. Technical Report Pb. No.:T/216. Forest Management Branch, Alberta Sustainable
    Resource Development, Edmonton, AB, CAN.  Appendix I. Pp. 21,22.
    '''
    if species == 'P' or \
               species == 'Pl' or \
               species == 'Pj' or \
               species == 'Pa' or \
               species == 'Pf':
        b1 = 12.84571
        b2 = -5.73936
        b3 = -0.91312
        b4 = 0.150668
        k1 = numpy.exp(b1+b2*(numpy.log(50+1)**0.5) + b3*numpy.log(si0)+b4*(50**0.5))
        k2 = si0**b3
        k3 = ((si0)*(1+k1)/1.3-1)/(numpy.exp(b1)*numpy.exp(b4*(50**0.5))*k2)
        y2bh = numpy.exp((numpy.log(k3)/b2)**2)-1
    if species == 'Sw' or \
               species == 'Se' or \
               species == 'S'  or \
               species == 'Fd' or \
               species == 'Fb' or \
               species == 'Fa':
        b1 = 12.14943
        b2 = -3.77051
        b3 = -0.28534
        b4 = 0.165483
        k1 = numpy.exp(b1+b2*(numpy.log(50**2+1)**0.5) + b3*(numpy.log(si0)**2)+b4*(50**0.5))
        k2 = si0**(b3*numpy.log(si0))
        k3 = ((si0)*(1+k1)/1.3-1)/(numpy.exp(b1)*numpy.exp(b4*(50**0.5))*k2)
        y2bh = (numpy.exp((numpy.log(k3)/b2)**2)-1)**0.5
    if species == 'Sb' or \
               species == 'Lt' or \
               species == 'La' or \
               species == 'Lw' or \
               species == 'L':
        b1 = 14.56236
        b2 = -6.04705
        b3 = -1.53715
        b4 = 0.240174
        k1 = numpy.exp(b1+b2*(numpy.log(50+1)**0.5) + b3*numpy.log(si0)+b4*(50**0.5))
        k2 = si0**b3
        k3 = ((si0)*(1+k1)/1.3-1)/(numpy.exp(b1)*numpy.exp(b4*(50**0.5))*k2)
        y2bh = numpy.exp((numpy.log(k3)/b2)**2)-1
    if species == 'Aw' or \
               species == 'Bw' or \
               species == 'Pb' or \
               species == 'A' or \
               species == 'H':
        b1 = 9.908888
        b2 = -3.92451
        b3 = -0.32778
        b4 = 0.134376
        k1 = numpy.exp(b1+b2*(numpy.log(50+1)**0.5) + b3*(numpy.log(si0)**2)+b4*(50**0.5))
        k2 = si0**(b3*numpy.log(si0))
        k3 = ((si0)*(1+k1)/1.3-1)/(numpy.exp(b1)*numpy.exp(b4*(50**0.5))*k2)
        y2bh = numpy.exp((numpy.log(k3)/b2)**2)-1
    return  y2bh



"""Stand density factor estimators"""
import logging

from pygypsy.density import (
    estimate_density_aw,
    estimate_density_sw,
    estimate_density_sb,
    estimate_density_pl,
)

LOGGER = logging.getLogger(__name__)


def estimate_sdf_aw(spc, site_index, bhage, density):
    '''Main purpose of this function is to estimate SDF for the species

    :param str spc: species_abbreviation
    :param float site_index: site index of species Aw
    :param float bhage: breast height age of speceis Aw
    :param float density: density of species Aw

    '''
    density_est = 0
    sdf = 0

    if density <= 0 or (bhage <= 0 and site_index <= 0):
        return density_est, sdf

    if spc not in ('Aw', 'Bw', 'Pb', 'A', 'H'):
        raise ValueError('%s is not a valid species', spc)

    sdf = density # best SDF guess
    tolerance = 0.00001
    within_tolerance = False
    iter_count = 0

    while not within_tolerance:
        result = estimate_density_aw(sdf, bhage, site_index, ret_detail=True)
        k1 = result['k1'] #pylint: disable=invalid-name
        k2 = result['k2'] #pylint: disable=invalid-name
        density_est = result['density']

        if abs(density-density_est) < tolerance:
            within_tolerance = True
        else:
            density_est = (density + density_est) / 2
            sdf = density_est * k2 / k1

        iter_count += 1

        if iter_count == 1500:
            LOGGER.warning('Slow convergence')
            break

    return density_est, sdf


def estimate_sdf_sb(spc, site_index, tage, density):
    '''Main purpose of this function is to estimate SDF for the species

    :param str spc: species_abbreviation
    :param float site_index: site index of species Sb
    :param float tage: total age of species Sb
    :param float density: density of species Sb

    '''
    density_est = 0
    sdf = 0

    if density <= 0 or (tage <= 0 and site_index <= 0):
        return density_est, sdf

    if spc not in ('Sb', 'Lt', 'La', 'Lw', 'L'):
        raise ValueError('%s is not a valid species', spc)

    sdf = density # best SDF guess
    tolerance = 0.00001
    within_tolerance = False
    iter_count = 0

    while not within_tolerance:
        result = estimate_density_sb(sdf, tage, site_index, ret_detail=True)
        k1 = result['k1'] #pylint: disable=invalid-name
        k2 = result['k2'] #pylint: disable=invalid-name
        density_est = result['density']

        if abs(density-density_est) < tolerance:
            within_tolerance = True
        else:
            density_est = (density + density_est)/2
            sdf = density_est * k2/k1

        iter_count += 1

        if iter_count == 150:
            LOGGER.warning('Slow convergence')
            break

    return density_est, sdf


def estimate_sdf_sw(spc, site_index, tage, sdfaw, density):
    '''Main purpose of this function is to estimate SDF for the species

    :param str spc: species_abbreviation
    :param float site_index: site index of species Sw
    :param float tage: total age of species Sw
    :param float sdfaw: Stand Density Factor of species Aw, this parameter indicates that
    the density of Sw depends on the density of Aw
    :param float density: density of species Sw

    '''
    density_est = 0
    sdf = 0

    if density <= 0 or (tage <= 0 and site_index <= 0):
        return density_est, sdf

    if spc not in ('Sw', 'Se', 'Fd', 'Fb', 'Fa'):
        raise ValueError('%s is not a valid species', spc)

    sdf = density # best SDF guess
    tolerance = 0.00001
    within_tolerance = False
    iter_count = 0

    while not within_tolerance:
        result = estimate_density_sw(sdf, sdfaw, tage, site_index, ret_detail=True)
        k1 = result['k1'] #pylint: disable=invalid-name
        k2 = result['k2'] #pylint: disable=invalid-name
        density_est = result['density']

        if abs(density-density_est) < tolerance:
            within_tolerance = True
        else:
            density_est = (density + density_est)/2
            sdf = density_est * k2/k1

        iter_count += 1

        if iter_count == 150:
            LOGGER.warning('Slow convergence')
            break

    return density_est, sdf


def estimate_sdf_pl(spc, site_index, tage, sdfaw, sdfsw, sdfsb, density):
    '''Main purpose of this function is to estimate SDF for the species

    :param str spc: species_abbreviation
    :param float site_index: site index of species Pl
    :param float tage: total age of species Pl
    :param float sdfaw: Stand Density Factor of species Aw
    :param float sdfsw: Stand Density Factor of species Sw
    :param float sdfpl: Stand Density Factor of species Pl
    these parameters SDF above indicate that the density of Pl
    depends on the density of all otehr species
    :param float density: density of species Pl

    '''
    density_est = 0
    sdf = 0

    if density <= 0 or (tage <= 0 and site_index <= 0):
        return density_est, sdf

    if spc not in ('P', 'Pl', 'Pj', 'Pa', 'Pf'):
        raise ValueError('%s is not a valid species', spc)

    sdf = density # best SDF guess
    tolerance = 0.00001
    within_tolerance = False
    iter_count = 0

    while not within_tolerance:
        result = estimate_density_pl(sdfaw, sdfsw, sdfsb, sdf,
                                     tage, site_index, ret_detail=True)
        k1 = result['k1'] #pylint: disable=invalid-name
        k2 = result['k2'] #pylint: disable=invalid-name
        density_est = result['density']

        if abs(density-density_est) < tolerance:
            within_tolerance = True
        else:
            density_est = (density + density_est)/2
            sdf = density_est * k2/k1

        iter_count += 1

        if iter_count == 150:
            LOGGER.warning('Slow convergence')
            break

    return density_est, sdf



from codecs import open as codecs_open
from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize
try:
    import numpy as np
except ImportError:
    msg = 'Numpy must be installed to install pygypsy'
    raise ImportError(msg)

import versioneer


with codecs_open('README.rst', encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

extensions = [
    Extension('pygypsy.basal_area_increment',
              ['pygypsy/basal_area_increment.pyx'],
              include_dirs = [np.get_include()]),
]


setup(name='pygypsy',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description=u'Forestry Growth and Yield Projection System',
      long_description=LONG_DESCRIPTION,
      keywords='',
      author=u'Julianno Sambatti, Jotham Apaloo, Ian Moss',
      author_email='julianno.sambatti@tesera.com',
      maintainer=u'Jotham Apaloo',
      maintainer_email=u'jotham.apaloo@tesera.com',
      url='',
      license='MIT',
      packages=find_packages(exclude=['tests']),
      ext_modules=cythonize(extensions),
      zip_safe=False,
      package_data={
          'pygypsy': ['scripts/config/*', '*.c', '*.pyx'],
      },
      install_requires=[
          'click>=6.6',
          'pandas>=0.18.1',
          'matplotlib>=1.5.2',
          'colorlog>=2.7.0',
          'jsonschema>=2.5.1',
          'boto3>=1.4.1',
          'boto>=2.43.0',
      ],
      extras_require={
          'test': ['pytest==2.9.1', 'pytest-cov==2.4.0'],
          'lint': ['pylint==1.5.4'],
          'docs': ['sphinx==1.4.1'],
          'dev': [
              'git-pylint-commit-hook==2.1.1',
              'Cython==0.25.1',
              'configparser==3.5.0', # compat for git-pylint-commit-hook
              'versioneer==0.17',
              'twine>=1.8.1',
          ],
          'analysis': [
              'jupyter>=1.0.0',
              'scikit-learn>=0.18',
              'snakeviz>=0.4.1',
          ],
      },
      entry_points='''
      [console_scripts]
      pygypsy=pygypsy.scripts.cli:cli
      ''',
      classifiers=[
          'Environment :: Console',
          'License :: OSI Approved :: MIT License',
          'Intended Audience :: End Users/Desktop',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 2 :: Only',
          'Topic :: Scientific/Engineering',
      ],
     )



"""CLI option callbacks"""
# this should probably not be tied to click exceptions
import json
import codecs
import logging

import click
import boto3
import jsonschema

from pygypsy.scripts import CONF_SCHEMA_FILE, DEFAULT_CONF_FILE
from pygypsy.log import CONSOLE_LOGGER_NAME
from pygypsy.utils import _parse_s3_url


LOGGER = logging.getLogger(CONSOLE_LOGGER_NAME)


def _load_and_validate_config(ctx, param, value): #pylint: disable=unused-argument, missing-docstring
    with open(CONF_SCHEMA_FILE) as schema_file:
        schema = json.load(schema_file)

    if value == DEFAULT_CONF_FILE:
        LOGGER.warning(
            'Using pygypsy default config file.'
        )

    try:
        if value.startswith('s3://'):
            s3_params = _parse_s3_url(value)
            client = boto3.client('s3')
            data = client.get_object(
                Bucket=s3_params['bucket'],
                Key=s3_params['prefix']
            )["Body"].read().decode('utf-8')
            conf = json.loads(data)
        else:
            with codecs.open(value, encoding='utf-8') as conf_file:
                conf = json.load(conf_file)
    except (IOError, ValueError) as err:
        LOGGER.error('Error reading config file: %s', value)
        raise click.BadParameter('Missing or incorrect filetype.')

    validator = jsonschema.Draft4Validator(schema)
    num_err = 0
    for error in sorted(validator.iter_errors(conf), key=str):
        msg = '/'.join(error.relative_path) + ': ' + error.message
        LOGGER.error(msg)
        num_err += 1

    if num_err > 0:
        raise click.BadParameter('Invalid config')

    return conf



"""Functions for calculating site indices

"""
from pygypsy.asaCompileAgeGivenSpSiHt import (
    ComputeGypsySiteIndex,
)


def get_site_indices_from_dominant_species(dominant_species=None,
                                           dominant_species_site_index=None):
    """Calculate site indices for non-dominant species

    Keyword Arguments:
    dominant_species            -- (default None)
    dominant_species_site_index -- (default None)

    """

    functs = {
        'aw': _get_all_site_indices_from_dominant_aw,
        'sw': _get_all_site_indices_from_dominant_sw,
        'fb': _get_all_site_indices_from_dominant_fb,
        'fd': _get_all_site_indices_from_dominant_fd,
        'pl': _get_all_site_indices_from_dominant_pl,
        'pb': _get_all_site_indices_from_dominant_pb,
    }
    dominant_species = dominant_species.lower()
    try:
        all_site_indices = functs[dominant_species](dominant_species_site_index)
    except KeyError:
        raise ValueError('No function is available to calculate site index from '
                         'species %s' %dominant_species)

    species_subset = ('Aw', 'Pl', 'Sw', 'Sb')
    site_indices_subset = {
        species: all_site_indices[species.lower()] for species in species_subset
    }

    return site_indices_subset


def _get_all_site_indices_from_dominant_aw(site_index_aw):
    """Keyword Arguments:
    site_index_aw -- site index for aw

    Returns:
    dict: species as keys and site indices as corresponding values

    """
    site_index_pl = 0.85 * site_index_aw + 3.4
    site_index_sw = 1.31 * site_index_aw - 2.64
    site_index_fb = 0.92 * site_index_sw + 1.68
    site_index_fd = 0.94 * site_index_pl + 0.71
    site_index_sb = 0.64 * site_index_pl + 2.76
    site_index_pb = site_index_aw

    site_indices_dict = {'aw': site_index_aw,
                         'pl': site_index_pl,
                         'sw': site_index_sw,
                         'fb': site_index_fb,
                         'fd': site_index_fd,
                         'sb': site_index_sb,
                         'pb': site_index_pb}

    return site_indices_dict


def _get_all_site_indices_from_dominant_sw(site_index_sw):
    """Keyword Arguments:
    site_index_sw -- site index for sw

    Returns:
    dict: species as keys and site indices as corresponding values

    """
    site_index_pl = 0.86 * site_index_sw + 2.13
    site_index_aw = 0.76 * site_index_sw + 2.01
    site_index_fb = 0.92 * site_index_sw + 1.68
    site_index_fd = 0.74 * site_index_sw + 4.75
    site_index_sb = 0.64 * site_index_pl + 2.76
    site_index_pb = site_index_aw

    site_indices_dict = {'aw': site_index_aw,
                         'pl': site_index_pl,
                         'sw': site_index_sw,
                         'fb': site_index_fb,
                         'fd': site_index_fd,
                         'sb': site_index_sb,
                         'pb': site_index_pb}

    return site_indices_dict

def _get_all_site_indices_from_dominant_fb(site_index_fb):
    """Keyword Arguments:
    site_index_fb -- site index for fb

    Returns: dict: species as keys and site indices as corresponding values

    """
    site_index_sw = 1.09 * site_index_fb - 1.83
    site_index_pl = 0.86 * site_index_sw + 2.13
    site_index_aw = 0.76 * site_index_sw + 2.01
    site_index_fd = 0.74 * site_index_sw + 4.75
    site_index_sb = 0.64 * site_index_pl + 2.76
    site_index_pb = site_index_aw

    site_indices_dict = {'aw': site_index_aw,
                         'pl': site_index_pl,
                         'sw': site_index_sw,
                         'fb': site_index_fb,
                         'fd': site_index_fd,
                         'sb': site_index_sb,
                         'pb': site_index_pb}

    return site_indices_dict


def _get_all_site_indices_from_dominant_fd(site_index_fd):
    """Keyword Arguments:
    site_index_fd -- site index for fd

    Returns:
    dict: species as keys and site indices as corresponding values

    """
    site_index_pl = 1.07 * site_index_fd - 0.76
    site_index_aw = 1.18 * site_index_pl  - 4.02
    site_index_sw = 1.36 * site_index_fd  - 6.45
    site_index_pb = site_index_aw
    site_index_fb = 0.92 * site_index_sw + 1.68
    site_index_sb = 0.64 * site_index_pl + 2.76

    site_indices_dict = {'aw': site_index_aw,
                         'pl': site_index_pl,
                         'sw': site_index_sw,
                         'fb': site_index_fb,
                         'fd': site_index_fd,
                         'sb': site_index_sb,
                         'pb': site_index_pb}

    return site_indices_dict


def _get_all_site_indices_from_dominant_pl(site_index_pl):
    """Keyword Arguments:
    site_index_pl -- site index for pl

    Returns:
    dict: species as keys and site indices as corresponding values

    """
    site_index_aw = 1.18 * site_index_pl  - 4.02
    site_index_sw = 1.16 * site_index_pl  - 2.47
    site_index_fd = 0.94* site_index_pl + 0.71
    site_index_pb = site_index_aw
    site_index_fb = 0.92 * site_index_sw + 1.68
    site_index_sb = 0.64 * site_index_pl + 2.76


    site_indices_dict = {'aw': site_index_aw,
                         'pl': site_index_pl,
                         'sw': site_index_sw,
                         'fb': site_index_fb,
                         'fd': site_index_fd,
                         'sb': site_index_sb,
                         'pb': site_index_pb}

    return site_indices_dict

def _get_all_site_indices_from_dominant_pb(site_index_pb):
    """Keyword Arguments:
    site_index_pb -- site index for pb

    Returns:
    dict: species as keys and site indices as corresponding values

    """
    site_index_aw = site_index_pb
    site_index_pl = 0.85 * site_index_aw + 3.4
    site_index_sw = 1.31 * site_index_aw - 2.64
    site_index_fd = 0.92 * site_index_pl + 1.68
    site_index_fb = 0.92 * site_index_sw + 1.68
    site_index_sb = 0.64 * site_index_pl + 2.76

    site_indices_dict = {'aw': site_index_aw,
                         'pl': site_index_pl,
                         'sw': site_index_sw,
                         'fb': site_index_fb,
                         'fd': site_index_fd,
                         'sb': site_index_sb,
                         'pb': site_index_pb}

    return site_indices_dict


def _get_temporary_dominant_species(actual_dominant_species):
    """Swap an actual dominant species for a temporary one
    Keyword Arguments:
    actual_dominant_species -- str

    Returns:
    str: a temporary dominant species

    """
    if actual_dominant_species == 'Pb':
        temp_dominant_species= 'Aw'
    elif actual_dominant_species == 'Fd' or actual_dominant_species == 'Fb':
        temp_dominant_species= 'Sw'
    else:
        raise ValueError('There is no temporary dominant species for species %s' \
                         % actual_dominant_species)

    return temp_dominant_species


def _estimate_site_index(species, age, height):
    """Estimate site index of dominant species

    This is a wrapper around ComputeGypsySiteIndex for readability

    Keyword Arguments:
    species -- str, abbreviation of dominant species
    age     -- float, age of dominant species
    height  -- float, height of dominant species

    Return:
    float - site index

    """
    dominant_site_index_list = ComputeGypsySiteIndex(species,
                                                     height, 0, age)
    dominant_site_index = dominant_site_index_list[2]

    return dominant_site_index



"""Package Loggers"""
import logging.config

CONSOLE_LOGGER_NAME = 'console'
FILE_LOGGER_NAME = 'file'

def setup_logging():
    """Initialize logging for pygypsy package"""
    log_conf = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'verbose': {
                'format': '%(levelname)s %(asctime)s %(process)d %(thread)d %(message)s'
            },
            'standard': {
                'format': '%(pathname)s: %(funcName)s %(levelname)s %(asctime)s %(message)s'
            },
            'colored': {
                '()': 'colorlog.ColoredFormatter',
                'format': "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(message)s"
            },
        },
        'handlers': {
            'console':{
                'level':'INFO',
                'class':'logging.StreamHandler',
                'formatter': 'colored'
            },
            'file': {
                'level': 'DEBUG',
                'formatter': 'standard',
                'class': 'logging.FileHandler',
                'filename': './pygypsy.log',
                'mode': 'w',
            },
        },
        'loggers': {
            'pygypsy.asaCompileAgeGivenSpSiHt': {
                'handlers':['file'],
                'level':'INFO',
                'propagate': True,
            },
            'pygypsy.forward_simulation': {
                'handlers':['file'],
                'level':'INFO',
                'propagate': True,
            },
            'pygypsy.data_prep': {
                'handlers':['file'],
                'level':'DEBUG',
                'propagate': True,
            },
            'pygypsy.utils': {
                'handlers':['file'],
                'level':'DEBUG',
                'propagate': True,
            },
            'pygypsy.basal_area_simulate': {
                'handlers':['file'],
                'level':'DEBUG',
                'propagate': True,
            },
            'pygypsy.basal_area_factor': {
                'handlers':['file'],
                'level':'DEBUG',
                'propagate': True,
            },
            'pygypsy.density': {
                'handlers':['file'],
                'level':'DEBUG',
                'propagate': True,
            },
            'pygypsy.plot': {
                'handlers':['file'],
                'level':'DEBUG',
                'propagate': True,
            },
            'pygypsy.stand_density_factor': {
                'handlers':['file'],
                'level':'DEBUG',
                'propagate': True,
            },
            'pygypsy.volume': {
                'handlers':['file'],
                'level':'DEBUG',
                'propagate': True,
            },
            'pygypsy.GYPSYNonSpatial': {
                'handlers':['file'],
                'level':'INFO',
                'propagate': True,
            },
            'pygypsy.site_index': {
                'handlers':['file'],
                'level':'DEBUG',
                'propagate': True,
            },
            CONSOLE_LOGGER_NAME: {
                'handlers': ['console', 'file'],
                'level': 'DEBUG',
                'propagate': True,
            },
        }
    }

    logging.config.dictConfig(log_conf)
